/********************************************************************************************
*
* Abstract: Elliptic curve arithmetic used for both point arithmetic and computing the
*           Weil pairing.
*
* Written by David Urbanik (dburbani@uwaterloo.ca) except for code apated from the original
* Microsoft implementation as indicated near the bottom.
*********************************************************************************************/ 

#include "definitions.h"
#include "constants.h"


inline void copy_point(full_point_proj_t P, full_point_proj_t Q) {
    fp2copy751_proj(P->X, Q->X);
    fp2copy751_proj(P->Y, Q->Y);
}

void tan_line_proj_mont(f2elm_t a, f2elm_t b_inv, full_point_proj_t P, f2elm_proj_t slope) {
    f2elm_proj_t tmp1, tmp2;

    fp2mul751_mont(P->Y->n, P->X->d, tmp1->n);
    fp2mul751_mont(P->Y->d, P->X->n, tmp1->d);                  //  tmp1 <- y/x

    fpmul751_mont(P->X->n[0], two_mont, tmp2->n[0]);
    fpmul751_mont(P->X->n[1], two_mont, tmp2->n[1]);
    fp2copy751(P->X->d, tmp2->d);
    fp2mixedadd751_mont(a, tmp2, tmp2);                         //  tmp2 <- 2x + a

    fp2mul751_mont(tmp2->n, tmp1->d, tmp2->n);
    fp2mul751_mont(tmp2->d, tmp1->n, tmp2->d);                  //  tmp2 <- (2x + a)(x/y)
    fp2mul751_mont(tmp2->n, b_inv, tmp2->n);
    fpmul751_mont(tmp2->d[0], two_mont, tmp2->d[0]);
    fpmul751_mont(tmp2->d[1], two_mont, tmp2->d[1]);            //  tmp2 <- x(2x + a)/(2by)

    fpmul751_mont(tmp1->d[0], two_mont, tmp1->d[0]);
    fpmul751_mont(tmp1->d[1], two_mont, tmp1->d[1]);            //  tmp1 <- y/(2x)

    fp2projadd751_mont(tmp1, tmp2, slope);
}

void sum_proj_mont(f2elm_t a, f2elm_t b_inv, full_point_proj_t P, full_point_proj_t Q, full_point_proj_t R) {
    f2elm_proj_t tmp1, tmp2, tmp3;

    //  compute tmp1 <- (y2 - y1)/(x2 - x1)
    fp2projsub751_mont(Q->Y, P->Y, tmp2);
    fp2projsub751_mont(Q->X, P->X, tmp3);
    fp2mul751_mont(tmp2->n, tmp3->d, tmp1->n);
    fp2mul751_mont(tmp2->d, tmp3->n, tmp1->d);

    //  compute tmp2 <- a + x1 + x2
    fp2projadd751_mont(P->X, Q->X, tmp2);
    fp2mixedadd751_mont(a, tmp2, tmp2);

    //  compute tmp3 <- b(tmp1)^2
    fp2projsqr751_mont(tmp1, tmp3);
    fp2mul751_mont(b_inv, tmp3->d, tmp3->d);

    //  compute tmp3 <- tmp3 - tmp2 = x(R)
    fp2projsub751_mont(tmp3, tmp2, tmp3);

    //  compute y(R) <- tmp1*(x(P) - x(R)) - y(P)
    fp2projsub751_mont(P->X, tmp3, tmp2);
    fp2projmul751_mont(tmp1, tmp2, tmp2);

    fp2projsub751_mont(tmp2, P->Y, R->Y);   //  store y(R)
    fp2copy751(tmp3->n, R->X->n);
    fp2copy751(tmp3->d, R->X->d);           //  store x(R)
}

void dbl_proj_mont(f2elm_t a, f2elm_t b_inv, full_point_proj_t P, full_point_proj_t P2, f2elm_proj_t l, int use_l) {
    f2elm_proj_t tmp1, tmp2;

    //  get l, the tangent line
    if (!use_l) {
        tan_line_proj_mont(a, b_inv, P, l);
    }

    //  compute tmp1 <- a + 2x
    fpmul751_mont(two_mont, P->X->n[0], tmp1->n[0]);
    fpmul751_mont(two_mont, P->X->n[1], tmp1->n[1]);
    fp2copy751(P->X->d, tmp1->d);                           //  tmp1 <- 2x
    fp2mixedadd751_mont(a, tmp1, tmp1);                     //  tmp1 <- a + 2x

    //  compute tmp2 <- b*l^2
    fp2projsqr751_mont(l, tmp2);                            //  tmp2 <- l^2
    fp2mul751_mont(b_inv, tmp2->d, tmp2->d);                //  tmp2 <- b * l^2

    //  compute x(P2) <- b*l^2 - (a + 2x) = tmp2 - tmp1
    fp2projsub751_mont(tmp2, tmp1, tmp2);                   //  tmp2 <- x(P2)

    //  We can now compute y(P2) as:
    //  y(P2) = -(y + (x(P2) - x)*l)
    fp2projsub751_mont(tmp2, P->X, tmp1);                   //  tmp1 <- x(P2) - x
    fp2projmul751_mont(tmp1, l, tmp1);                      //  tmp1 <- (x(P2) - x)*l
    fp2projadd751_mont(P->Y, tmp1, P2->Y);                  //  P2->Y <- -y(P2)

    fp2neg751(P2->Y->n);                                    //  P2->Y <- y(P2)
    fp2copy751(tmp2->n, P2->X->n);
    fp2copy751(tmp2->d, P2->X->d);                          //  P2->X <- x(P2)
}

int is_inf(full_point_proj_t P) {
    return (is_f2elm_eq(P->Y->d, f2_zero));
}

void set_inf(full_point_proj_t P) {
    fp2copy751(f2_zero, P->Y->d);
}

void get_y_sqred_mont(f2elm_t x, f2elm_t a, f2elm_t b_inv, f2elm_t y_sqred) {
    f2elm_t tmp;

    fp2sqr751_mont(x, tmp);                     //  tmp     <- x^2
    fp2mul751_mont(tmp, a, y_sqred);            //  y_sqred <- a*x^2
    fp2add751(y_sqred, x, y_sqred);             //  y_sqred <- a*x^2 + x
    fp2mul751_mont(tmp, x, tmp);                //  tmp     <- x^3
    fp2add751(y_sqred, tmp, y_sqred);           //  y_sqred <- x^3 + a*x^2 + x
    fp2mul751_mont(y_sqred, b_inv, y_sqred);    //  y_sqred <- b^-1 * (x^3 + a*x^2 + x) = y^2
}

void get_y_sqred_proj_mont(f2elm_proj_t x, f2elm_t a, f2elm_t b_inv, f2elm_proj_t y_sqred) {
    f2elm_proj_t tmp;
    fp2projsqr751_mont(x, tmp);                     //  tmp     <- x^2
    fp2mixedmul751_mont(a, tmp, y_sqred);           //  y_sqred <- a*x^2
    fp2projadd751_mont(y_sqred, x, y_sqred);        //  y_sqred <- a*x^2 + x
    fp2projmul751_mont(tmp, x, tmp);                //  tmp     <- x^2
    fp2projadd751_mont(y_sqred, tmp, y_sqred);      //  y_sqred <- x^3 + a*x^2 + x
    fp2mixedmul751_mont(b_inv, y_sqred, y_sqred);   //  y_sqred <- b^-1 * (x^3 + a*x^2 + x) = y^2
}

int is_on_curve_mont(f2elm_t a, f2elm_t b_inv, f2elm_t x, f2elm_t y) {
    f2elm_t y_sqred_from_x, y_sqred;
    get_y_sqred_mont(x, a, b_inv, y_sqred_from_x);
    fp2sqr751_mont(y, y_sqred);

    return is_f2elm_eq(y_sqred, y_sqred_from_x);
}

void add_points(f2elm_t a, f2elm_t b_inv, full_point_proj_t P, full_point_proj_t Q, full_point_proj_t R, f2elm_proj_t l, int use_l) {
    int i;

    if (is_inf(P)) {
        copy_point(Q, R);
    } else if (is_inf(Q)) {
        copy_point(P, R);
    } else if (is_f2elm_proj_eq(P->X, Q->X)) {
        if (is_f2elm_proj_eq(P->Y, Q->Y)) {
            //  we have a doubling operation
            if (is_f2elm_eq(P->Y->n, f2_zero)) {
                set_inf(R);
            } else {
                dbl_proj_mont(a, b_inv, P, R, l, use_l);
            }
        } else {
            set_inf(R);
        }
    } else {
        //  we have a normal addition
        sum_proj_mont(a, b_inv, P, Q, R);
    }
}

void add_points_generic(f2elm_t a, f2elm_t b_inv, full_point_proj_t P, full_point_proj_t Q, full_point_proj_t R) {
    f2elm_proj_t dummy;
    add_points(a, b_inv, P, Q, R, dummy, 0);
}

void h_TU_mont(full_point_proj_t T, full_point_proj_t U, full_point_proj_t S, f2elm_proj_t result) {
    f2elm_proj_t tmp1, tmp2, tmp3;

    if (is_f2elm_proj_eq(T->X, U->X)) {
        fp2projsub751_mont(S->X, T->X, result);        
    } else {
        fp2projsub751_mont(T->X, U->X, tmp1);       //  tmp1 <- (xT - xU)
        fp2projsub751_mont(T->Y, U->Y, tmp2);       //  tmp2 <- (yT - yU)
        fp2mul751_mont(tmp2->n, tmp1->d, tmp2->n);
        fp2mul751_mont(tmp2->d, tmp1->n, tmp2->d);  //  tmp2 <- (yT - yU)/(xT - xU)

        fp2projsub751_mont(T->Y, S->Y, tmp1);       //  tmp1 <- yT - yS
        fp2projsub751_mont(S->X, T->X, tmp3);       //  tmp3 <- xS - xT
        fp2projmul751_mont(tmp3, tmp2, tmp2);       //  tmp2 <- (xS - xT)(yT - yU)/(xT - xU)
        fp2projadd751_mont(tmp1, tmp2, result);
    }
}

void multi_h_TU_mont(full_point_proj_t T, full_point_proj_t U, full_point_proj_t *S, f2elm_proj_t *result, int eval_num) {
    int i;
    f2elm_proj_t tmp1, tmp2, tmp3;

    if (is_f2elm_proj_eq(T->X, U->X)) {
        for (i = 0; i < eval_num; i++) {
            fp2projsub751_mont(S[i]->X, T->X, result[i]);
        }
    } else {
        fp2projsub751_mont(T->X, U->X, tmp1);       //  tmp1 <- (xT - xU)
        fp2projsub751_mont(T->Y, U->Y, tmp2);       //  tmp2 <- (yT - yU)
        fp2mul751_mont(tmp2->n, tmp1->d, tmp2->n);
        fp2mul751_mont(tmp2->d, tmp1->n, tmp2->d);  //  tmp2 <- (yT - yU)/(xT - xU)

        for (i = 0; i < eval_num; i++) {
            fp2projsub751_mont(T->Y, S[i]->Y, tmp1);       //  tmp1 <- yT - yS
            fp2projsub751_mont(S[i]->X, T->X, tmp3);       //  tmp3 <- xS - xT
            fp2projmul751_mont(tmp3, tmp2, tmp3);       //  tmp3 <- (xS - xT)(yT - yU)/(xT - xU)
            fp2projadd751_mont(tmp1, tmp3, result[i]);
        }
    }
}

void h_TT_mont(f2elm_t a, f2elm_t b_inv, full_point_proj_t T, full_point_proj_t S, f2elm_proj_t result, f2elm_proj_t l, int *computed_l) {
    f2elm_proj_t tmp1, tmp2;

    if (!(*computed_l)) {
        tan_line_proj_mont(a, b_inv, T, l);
        *computed_l = 1;
    }

    if (is_f2elm_eq(l->d, f2_zero)) {
        fp2projsub751_mont(S->X, T->X, result);
    } else {
        fp2projsub751_mont(T->Y, S->Y, tmp1);       //  tmp1 <- yT - yS
        fp2projsub751_mont(S->X, T->X, tmp2);       //  tmp2 <- xS - xT
        fp2projmul751_mont(tmp2, l, tmp2);          //  tmp2 <- l*(xS - xT)
        fp2projadd751_mont(tmp1, tmp2, result); 
    }
}

void h_S_mont(full_point_proj_t T, full_point_proj_t S, f2elm_proj_t result) {
    if (is_inf(T)) {
        fp2copy751(f2_one_mont, result->n);
        fp2copy751(f2_one_mont, result->d);
    } else {
        fp2projsub751_mont(S->X, T->X, result);                  //  result <- xT - xS
    }
}

void multi_h_line_mont(f2elm_t a, f2elm_t b_inv, full_point_proj_t T, full_point_proj_t U, full_point_proj_t *S,
                       f2elm_proj_t *result, int eval_num, f2elm_proj_t l, int *computed_l) {
    int i;

    if (is_inf(T)) {
        for (i = 0; i < eval_num; i++) {
            h_S_mont(U, S[i], result[i]);
        }
    } else if (is_inf(U)) {
        for (i = 0; i < eval_num; i++) {
            h_S_mont(T, S[i], result[i]);
        }
    } else if (is_f2elm_proj_eq(T->X, U->X) && is_f2elm_proj_eq(T->Y, U->Y)) {
        for (i = 0; i < eval_num; i++) {
            h_TT_mont(a, b_inv, T, S[i], result[i], l, computed_l);
        }
    } else {
        multi_h_TU_mont(T, U, S, result, eval_num);
    }
}

void h_line_mont(f2elm_t a, f2elm_t b_inv, full_point_proj_t T, full_point_proj_t U, full_point_proj_t S, f2elm_proj_t result, f2elm_proj_t l, int *computed_l) {
    if (is_inf(T)) {
        h_S_mont(U, S, result);
    } else if (is_inf(U)) {
        h_S_mont(T, S, result);
    } else if (is_f2elm_proj_eq(T->X, U->X) && is_f2elm_proj_eq(T->Y, U->Y)) {
        h_TT_mont(a, b_inv, T, S, result, l, computed_l);
    } else {
        h_TU_mont(T, U, S, result);
    }
}

void g_factor(f2elm_t a, f2elm_t b_inv, full_point_proj_t T, full_point_proj_t U, full_point_proj_t TpU, full_point_proj_t S, f2elm_proj_t result) {
    f2elm_proj_t tmp1, l;
    int computed_l = 0;
    
    h_line_mont(a, b_inv, T, U, S, result, l, &computed_l);
    add_points(a, b_inv, T, U, TpU, l, computed_l);
    h_S_mont(TpU, S, tmp1);

    fp2mul751_mont(result->n, tmp1->d, result->n);
    fp2mul751_mont(result->d, tmp1->n, result->d);
}

void multi_g_factor(f2elm_t a, f2elm_t b_inv, full_point_proj_t T, full_point_proj_t U, full_point_proj_t TpU, full_point_proj_t *S, f2elm_proj_t *result, int eval_num) {
    f2elm_proj_t tmp, l;
    int computed_l = 0;
    int i;

    multi_h_line_mont(a, b_inv, T, U, S, result, eval_num, l, &computed_l);
    add_points(a, b_inv, T, U, TpU, l, computed_l);

    for (i = 0; i < eval_num; i++) {
        h_S_mont(TpU, S[i], tmp);
        fp2mul751_mont(result[i]->n, tmp->d, result[i]->n);
        fp2mul751_mont(result[i]->d, tmp->n, result[i]->d);
    }
}

void get_y_coord_mont(f2elm_t x, f2elm_t a, f2elm_t b_inv, f2elm_t y, int *on_curve) {
    //  Tests to see if there exists a point (x,y) on   E: b*y = x^3 + a*x^2 + x

    f2elm_t y_sqred;
    get_y_sqred_mont(x, a, b_inv, y_sqred);
    try_fp2sqrt751_mont(y_sqred, y, on_curve);
}

void get_affine(f2elm_proj_t f_proj, f2elm_t f_affine) {
    f2elm_t tmp_inv;
    fp2copy751(f_proj->d, tmp_inv);
    fp2inv751_mont(tmp_inv);
    fp2mul751_mont(f_proj->n, tmp_inv, f_affine);
}

void get_y2_mont(f2elm_proj_t x1, f2elm_proj_t x2, f2elm_proj_t x3, f2elm_proj_t y1, f2elm_t a, f2elm_t b_inv, f2elm_proj_t y2) {
    f2elm_proj_t y2_sqred, tmp1, tmp2, tmp3;

    get_y_sqred_proj_mont(x2, a, b_inv, y2_sqred);  //  y2_sqred <- y2^2
    fp2projsqr751_mont(y1, tmp1);                   //  tmp1     <- y1^2
    fp2projadd751_mont(tmp1, y2_sqred, tmp1);       //  tmp1     <- y1^2 + y2^2
    
    fp2projadd751_mont(x1, x2, tmp2);               //  tmp2     <- x1 + x2
    fp2projadd751_mont(tmp2, x3, tmp2);             //  tmp2     <- x1 + x2 + x3
    fp2mixedadd751_mont(a, tmp2, tmp2);             //  tmp2     <- x1 + x2 + x3 + a
    fp2mixedmul751_mont(b_inv, tmp2, tmp2);         //  tmp2     <- b^-1 * (x1 + x2 + x3 + a)

    fp2projsub751_mont(x1, x2, tmp3);               //  tmp3     <- x1 - x2
    fp2projsqr751_mont(tmp3, tmp3);                 //  tmp3     <- (x1 - x2)^2
    fp2projmul751_mont(tmp2, tmp3, tmp2);           //  tmp2     <- b^-1 * (x1 + x2 + x3 + a) * (x1 - x2)^2

    fp2projsub751_mont(tmp1, tmp2, tmp1);           //  tmp1     <- y1^2 + y2^2 - tmp2
    fp2copy751(y1->d, tmp3->d);
    fpmul751_mont(two_mont, y1->n[0], tmp3->n[0]);  //  tmp3[0]  <- y1[0] * 2
    fpmul751_mont(two_mont, y1->n[1], tmp3->n[1]);  //  tmp3[1]  <- y1[1] * 2

    fp2mul751_mont(tmp1->n, tmp3->d, y2->n);
    fp2mul751_mont(tmp1->d, tmp3->n, y2->d);
}

void get_two_tor_point_mont(f2elm_t x_rand, f2elm_t a, f2elm_t a24, f2elm_t b_inv, full_point_proj_t P) {
    int on_curve;
    f2elm_proj_t xQ_proj, x_rand_proj, y_rand_proj;
    f2elm_t xQ, y_rand, tmp;

    do {
        fp2copy751(x_rand, tmp);
        fp2sqr751_mont(x_rand, x_rand);
        fp2sqr751_mont(x_rand, x_rand);
        fp2mul751_mont(x_rand, tmp, x_rand);
        get_y_coord_mont(x_rand, a, b_inv, y_rand, &on_curve);
    } while (!on_curve);

    fp2copy751(x_rand, x_rand_proj->n);
    fp2copy751(f2_one_mont, x_rand_proj->d);
    fp2copy751(y_rand, y_rand_proj->n);
    fp2copy751(f2_one_mont, y_rand_proj->d);

    ladder(x_rand_proj, three_exp_239, P->X, xQ_proj, a24, 379, 384);
    get_y2_mont(x_rand_proj, P->X, xQ_proj, y_rand_proj, a, b_inv, P->Y);
}

void get_three_tor_point_mont(f2elm_t x_rand, f2elm_t a, f2elm_t a24, f2elm_t b_inv, full_point_proj_t P) {
    int on_curve;
    f2elm_proj_t xQ_proj, x_rand_proj, y_rand_proj;
    f2elm_t xQ, y_rand;

    do {
        fpadd751(x_rand[0], one_mont, x_rand[0]);
        fpadd751(x_rand[1], one_mont, x_rand[1]);
        fpadd751(x_rand[1], one_mont, x_rand[1]);
        get_y_coord_mont(x_rand, a, b_inv, y_rand, &on_curve);
    } while (!on_curve);

    fp2copy751(x_rand, x_rand_proj->n);
    fp2copy751(f2_one_mont, x_rand_proj->d);
    fp2copy751(y_rand, y_rand_proj->n);
    fp2copy751(f2_one_mont, y_rand_proj->d);

    ladder(x_rand_proj, two_exp_372, P->X, xQ_proj, a24, 373, 384);           
    get_y2_mont(x_rand_proj, P->X, xQ_proj, y_rand_proj, a, b_inv, P->Y);
}

void get_full_two_tor_point_mont(f2elm_t x_rand, f2elm_t a, f2elm_t a24, f2elm_t b_inv, full_point_proj_t P, f2elm_proj_t xR_proj) {
    int count = 0;
    do {
        get_two_tor_point_mont(x_rand, a, a24, b_inv, P);
        xonly_dbl_e(P->X, xR_proj, a, f2_one_mont, 371);
        count++;
    } while (is_f2elm_eq(xR_proj->d, f2_zero));
}

void get_full_three_tor_point_mont(f2elm_t x_rand, f2elm_t a, f2elm_t a24, f2elm_t b_inv, full_point_proj_t P, f2elm_proj_t xR_proj) {
    do {
        get_three_tor_point_mont(x_rand, a, a24, b_inv, P);
        xonly_tpl_e(P->X, xR_proj, a, f2_one_mont, 238);
    } while (is_f2elm_eq(xR_proj->d, f2_zero));
}

void get_generators_three_tor(f2elm_t a, f2elm_t a24, f2elm_t b_inv, full_point_proj_t Q1, full_point_proj_t Q2) {
    f2elm_proj_t dummy1, dummy2, dummy3;
    f2elm_t x_rand;
    fp2copy751(f2_one_mont, x_rand);

    get_full_three_tor_point_mont(x_rand, a, a24, b_inv, Q1, dummy1);
    xonly_dbl(dummy1, dummy2, a24, f2_one_mont);

    do {
        get_full_three_tor_point_mont(x_rand, a, a24, b_inv, Q2, dummy3);
    } while (is_f2elm_proj_eq(dummy1, dummy3) || is_f2elm_proj_eq(dummy2, dummy3)); 
}

void get_generators_two_tor(f2elm_t a, f2elm_t a24, f2elm_t b_inv, full_point_proj_t Q1, full_point_proj_t Q2) {
    f2elm_proj_t dummy1, dummy2;
    f2elm_t x_rand;
    fpcopy751(two_mont, x_rand[0]);
    fpcopy751(three_mont, x_rand[1]);

    get_full_two_tor_point_mont(x_rand, a, a24, b_inv, Q1, dummy1);   

    do {
        get_full_two_tor_point_mont(x_rand, a, a24, b_inv, Q2, dummy2); 
    } while (is_f2elm_proj_eq(dummy1, dummy2)); 
}


void multi_eval_f_miller(digit_t *subgp_order, int ord_len, f2elm_t a, f2elm_t b_inv, full_point_proj_t P, full_point_proj_t *S, f2elm_proj_t *result, int eval_num) {
    f2elm_proj_t tmp1[eval_num];
    full_point_proj_t Z;
    int i;

    for (i = 0; i < eval_num; i++) {
        fp2copy751(f2_one_mont, result[i]->n);
        fp2copy751(f2_one_mont, result[i]->d);
    }

    copy_point(P, Z);

    int byte_num = (ord_len-1)/(sizeof(digit_t)*8);
    int leading_bit = (ord_len % (sizeof(digit_t)*8));
    digit_t mask;

    if (leading_bit != 0) {
        mask = (((digit_t) 0x1) << (leading_bit-1));
    } else {
        mask = (((digit_t) 0x1) << (sizeof(digit_t)*8 - 1));
    }

    do {
        do {
            for (i = 0; i < eval_num; i++) {
                fp2projsqr751_mont(result[i], result[i]);
            }

            multi_g_factor(a, b_inv, Z, Z, Z, S, tmp1, eval_num);

            for (i = 0; i < eval_num; i++) {
                fp2projmul751_mont(result[i], tmp1[i], result[i]);
            }


            if (mask & subgp_order[byte_num]) {
                multi_g_factor(a, b_inv, Z, P, Z, S, tmp1, eval_num);
                for (i = 0; i < eval_num; i++) {
                    fp2projmul751_mont(result[i], tmp1[i], result[i]);
                }
            }

            mask = (mask >> 1);
        } while (mask != 0);

        mask = (((digit_t) 0x1) << (sizeof(digit_t)*8 - 1));
        byte_num--;
    } while (byte_num >= 0);
}


void eval_f_miller(digit_t *subgp_order, int ord_len, f2elm_t a, f2elm_t b_inv, full_point_proj_t P, full_point_proj_t S, f2elm_proj_t result) {
    f2elm_proj_t tmp1;
    full_point_proj_t Z;

    fp2copy751(f2_one_mont, result->n);
    fp2copy751(f2_one_mont, result->d);

    copy_point(P, Z);

    int byte_num = (ord_len-1)/(sizeof(digit_t)*8);
    int leading_bit = (ord_len % (sizeof(digit_t)*8));
    digit_t mask;

    if (leading_bit != 0) {
        mask = (((digit_t) 0x1) << (leading_bit-1));
    } else {
        mask = (((digit_t) 0x1) << (sizeof(digit_t)*8 - 1));
    }

    do {
        do {
            fp2projsqr751_mont(result, result);

            g_factor(a, b_inv, Z, Z, Z, S, tmp1);
            fp2projmul751_mont(result, tmp1, result);

            if (mask & subgp_order[byte_num]) {
                g_factor(a, b_inv, Z, P, Z, S, tmp1);
                fp2projmul751_mont(result, tmp1, result);
            }

            mask = (mask >> 1);
        } while (mask != 0);

        mask = (((digit_t) 0x1) << (sizeof(digit_t)*8 - 1));
        byte_num--;
    } while (byte_num >= 0);
}

void weil_pairing_three_tor(f2elm_t a, f2elm_t b_inv, full_point_proj_t P, full_point_proj_t Q, f2elm_t result) {
    f2elm_proj_t tmp1, tmp2;

    if (is_f2elm_proj_eq(P->X, Q->X)) {
        fp2copy751(f2_one_mont, result);
        return;
    }

    eval_f_miller(three_exp_239, 378, a, b_inv, P, Q, tmp1);
    eval_f_miller(three_exp_239, 378, a, b_inv, Q, P, tmp2);

    fp2mul751_mont(tmp1->n, tmp2->d, tmp1->n);
    fp2mul751_mont(tmp1->d, tmp2->n, tmp1->d);

    fp2inv751_mont(tmp1->d);
    fp2mul751_mont(tmp1->n, tmp1->d, result);

    fp2neg751(result);
}

void weil_pairing_two_tor(f2elm_t a, f2elm_t b_inv, full_point_proj_t P, full_point_proj_t Q, f2elm_t result) {
    f2elm_proj_t tmp1, tmp2;

    if (is_f2elm_proj_eq(P->X, Q->X)) {
        fp2copy751(f2_one_mont, result);
        return;
    }

    eval_f_miller(two_exp_372, 372, a, b_inv, P, Q, tmp1);
    eval_f_miller(two_exp_372, 372, a, b_inv, Q, P, tmp2);

    fp2mul751_mont(tmp1->n, tmp2->d, tmp1->n);
    fp2mul751_mont(tmp1->d, tmp2->n, tmp1->d);

    fp2inv751_mont(tmp1->d);
    fp2mul751_mont(tmp1->n, tmp1->d, result);
}

/***************************************************************************************************************
** The functions after this point are adapted from the original implementation by Microsoft in the SIDH library.
****************************************************************************************************************/

void xonly_dbl_add(f2elm_proj_t xP, f2elm_proj_t xQ, f2elm_proj_t xPQ, f2elm_t A24)
{ // Simultaneous doubling and differential addition.
  // Input: projective Montgomery points P=(nP:dP) and Q=(nQ:dQ) such that xP=nP/dP and xQ=nQ/dQ, affine difference xPQ=x(P-Q) and Montgomery curve constant A24=(A+2)/4.
  // Output: projective Montgomery points P <- 2*P = (n2P:d2P) such that x(2P)=n2P/d2P, and Q <- P+Q = (nQP:dQP) such that = x(Q+P)=nQP/dQP. 
    f2elm_t t0, t1, t2;

    fp2add751(xP->n, xP->d, t0);                        // t0 <- nP+dP
    fp2sub751(xP->n, xP->d, t1);                        // t1 <- nP-dP
    fp2sqr751_mont(t0, xP->n);                          // nP <- (nP+dP)^2
    fp2sub751(xQ->n, xQ->d, t2);                        // t2 <- nQ-dQ
    fp2add751(xQ->n, xQ->d, xQ->n);                     // nQ <- nQ+dQ
    fp2mul751_mont(t0, t2, t0);                         // t0 <- (nP+dP)*(nQ-dQ)
    fp2sqr751_mont(t1, xP->d);                          // dP <- (nP-dP)^2
    fp2mul751_mont(t1, xQ->n, t1);                      // t1 <- (nP-dP)*(nQ+dQ)
    fp2sub751(xP->n, xP->d, t2);                        // t2 <- (nP+dP)^2-(nP-dP)^2
    fp2mul751_mont(xP->n, xP->d, xP->n);                // nP <- (nP+dP)^2*(nP-dP)^2
    fp2mul751_mont(t2, A24, xQ->n);                     // nQ <- A24*[(nP+dP)^2-(nP-dP)^2]
    fp2sub751(t0, t1, xQ->d);                           // dQ <- (nP+dP)*(nQ-dQ)-(nP-dP)*(nQ+dQ)
    fp2add751(xQ->n, xP->d, xP->d);                     // dP <- A24*[(nP+dP)^2-(nP-dP)^2]+(nP-dP)^2
    fp2add751(t0, t1, xQ->n);                           // nQ <- (XP+ZP)*(nQ-dQ)+(nP-dP)*(nQ+dQ)
    fp2mul751_mont(xP->d, t2, xP->d);                   // dP <- [A24*[(nP+dP)^2-(nP-dP)^2]+(nP-dP)^2]*[(nP+dP)^2-(nP-dP)^2]
    fp2sqr751_mont(xQ->d, xQ->d);                       // dQ <- [(nP+dP)*(nQ-dQ)-(nP-dP)*(nQ+dQ)]^2
    fp2sqr751_mont(xQ->n, xQ->n);                       // nQ <- [(nP+dP)*(nQ-dQ)+(nP-dP)*(nQ+dQ)]^2
    
    fp2mul751_mont(xQ->d, xPQ->n, xQ->d);               
    fp2mul751_mont(xQ->n, xPQ->d, xQ->n);				// Q <- xPQ*[(nP+dP)*(nQ-dQ)-(nP-dP)*(nQ+dQ)]^2
}


void xonly_dbl(f2elm_proj_t xP, f2elm_proj_t xQ, f2elm_t A24, f2elm_t C24)
{ // Doubling of a Montgomery point in projective coordinates (n:d).
  // Input: projective Montgomery x-coordinates P = (n1:d1), where x1=n1/d1 and Montgomery curve constant A24/C24=(A/C+2)/4.
  // Output: projective Montgomery x-coordinates Q = 2*P = (n2:d2).
    f2elm_t t0, t1;
    
    fp2sub751(xP->n, xP->d, t0);                        // t0 <- n1-d1
    fp2add751(xP->n, xP->d, t1);                        // t1 <- n1+d1
    fp2sqr751_mont(t0, t0);                             // t0 <- (n1-d1)^2 
    fp2sqr751_mont(t1, t1);                             // t1 <- (n1+d1)^2 
    fp2mul751_mont(C24, t0, xQ->d);                     // d2 <- C24*(n1-d1)^2   
    fp2mul751_mont(t1, xQ->d, xQ->n);                   // n2 <- C24*(n1-d1)^2*(n1+d1)^2
    fp2sub751(t1, t0, t1);                              // t1 <- (n1+d1)^2-(n1-d1)^2 
    fp2mul751_mont(A24, t1, t0);                        // t0 <- A24*[(n1+d1)^2-(n1-d1)^2]
    fp2add751(xQ->d, t0, xQ->d);                        // d2 <- A24*[(n1+d1)^2-(n1-d1)^2] + C24*(n1-d1)^2
    fp2mul751_mont(xQ->d, t1, xQ->d);                   // d2 <- [A24*[(n1+d1)^2-(n1-d1)^2] + C24*(n1-d1)^2]*[(n1+d1)^2-(n1-d1)^2]
}


void xonly_dbl_e(f2elm_proj_t xP, f2elm_proj_t xQ, f2elm_t A, f2elm_t C, int e)
{ // Computes [2^e](n:d) on Montgomery curve with projective constant via e repeated doublings.
  // Input: projective Montgomery x-coordinates P = (nP:dP), such that xP=nP/dP and Montgomery curve constant A/C.
  // Output: projective Montgomery x-coordinates P <- (2^e)*P.
    f2elm_t A24num, A24den;
    int i;
    
    fp2add751(C, C, A24num);                           
    fp2add751(A24num, A24num, A24den);                    
    fp2add751(A24num, A, A24num);
    fp2copy751(xP->n, xQ->n);
    fp2copy751(xP->d, xQ->d);

    for (i = 0; i < e; i++) {
        xonly_dbl(xQ, xQ, A24num, A24den);
    }
}


void xonly_tpl(f2elm_proj_t xP, f2elm_proj_t xQ, f2elm_t A, f2elm_t C)
{ // Tripling of a Montgomery point in projective coordinates (n:d).
  // Input: projective Montgomery x-coordinates P = (n:d), where x=n/d and Montgomery curve constant A4=4*A.
  // Output: projective Montgomery x-coordinates Q = 3*P = (n3:n3).
    f2elm_t t0, t1, t2, t3, t4, t5;
    
    fp2add751(xP->n, xP->d, t2);                        // t2 <- n+d
    fp2sqr751_mont(xP->n, t0);                          // t0 <- n^2
    fp2sqr751_mont(xP->d, t1);                          // t1 <- d^2
    fp2sqr751_mont(t2, t2);                             // t2 <- (n+d)^2
    fp2mul751_mont(t0, C, t3);                          // t3 <- C*n^2
    fp2sub751(t2, t0, t2);                              // t2 <- (n+d)^2-n^2
    fp2mul751_mont(t1, C, t4);                          // t4 <- C*d^2
    fp2sub751(t2, t1, t2);                              // t2 <- 2nd = (n+d)^2-n^2-d^2
    fp2add751(t3, t4, t5);                              // t5 <- C*n^2+C*d^2
    fp2mul751_mont(t2, A, t2);                          // t2 <- 2And
    fp2add751(t3, t3, t3);                              // t3 <- 2C*n^2 
    fp2add751(t4, t4, t4);                              // t4 <- 2C*d^2
    fp2add751(t3, t2, t3);                              // t3 <- 2C*n^2+2And
    fp2add751(t4, t2, t4);                              // t4 <- 2C*d^2+2And
    fp2add751(t3, t5, t3);                              // t3 <- 2C*n^2+2And + C*n^2+C*d^2
    fp2add751(t4, t5, t4);                              // t4 <- 2C*d^2+2And + C*n^2+C*d^2
    fp2sub751(t0, t1, t2);                              // t2 <- n^2-d^2
    fp2add751(t0, t0, t0);                              // t0 <- 2n^2
    fp2add751(t1, t1, t1);                              // t1 <- 2d^2
    fp2mul751_mont(t2, t5, t2);                         // t2 <- (n^2-d^2)(C*n^2+C*d^2)
    fp2mul751_mont(t1, t3, t1);                         // t1 <- 2d^2*[2C*n^2+2And+C*n^2+C*d^2]
    fp2mul751_mont(t0, t4, t0);                         // t0 <- 2n^2*[2C*d^2+2And+C*n^2+C*d^2]
    fp2sub751(t1, t2, t1);                              // t1 <- 2d^2*[2C*n^2+2And+C*n^2+C*d^2] - (n^2-d^2)(C*n^2+C*d^2)
    fp2add751(t0, t2, t0);                              // t0 <- 2n^2*[2C*d^2+2And+C*n^2+C*d^2] + (n^2-d^2)(C*n^2+C*d^2)
    fp2sqr751_mont(t1, t1);                             // t1 <- [2d^2*[2C*n^2+2And+C*n^2+C*d^2] - (n^2-d^2)(C*n^2+C*d^2)]^2
    fp2sqr751_mont(t0, t0);                             // t0 <- [2n^2*[2C*d^2+2And+C*n^2+C*d^2] + (n^2-d^2)(C*n^2+C*d^2)]^2
    fp2mul751_mont(xP->n, t1, xQ->n);                   // n3 <- n*[2d^2*[2C*n^2+2And+C*n^2+C*d^2] - (n^2-d^2)(C*n^2+C*d^2)]^2
    fp2mul751_mont(xP->d, t0, xQ->d);                   // d3 <- d*[2n^2*[2C*d^2+2And+C*n^2+C*d^2] + (n^2-d^2)(C*n^2+C*d^2)]^2
}


void xonly_tpl_e(f2elm_proj_t xP, f2elm_proj_t xQ, f2elm_t A, f2elm_t C, int e)
{ // Computes [3^e](n:d) on Montgomery curve with projective constant via e repeated triplings.
  // Input: projective Montgomery x-coordinates P = (nP:dP), such that xP=nP/dP and Montgomery curve constant A/C.
  // Output: projective Montgomery x-coordinates P <- (3^e)*P.
    int i;
      
    copy_words((digit_t*)xP, (digit_t*)xQ, 2*2*NWORDS_FIELD);

    for (i = 0; i < e; i++) {
        xonly_tpl(xQ, xQ, A, C);
    }
}


void ladder(f2elm_proj_t xR, digit_t* m, f2elm_proj_t xP, f2elm_proj_t xQ, f2elm_t A24, unsigned int order_bits, unsigned int order_fullbits) {
    unsigned int bit = 0, owords = NBITS_TO_NWORDS(order_fullbits);
    digit_t mask;
    int i;

    fp2copy751(f2_one_mont, xP->n);
    fp2copy751(f2_zero, xP->d);

    fp2copy751_proj(xR, xQ);

    digit_t m_cpy[order_fullbits/(8*sizeof(digit_t))];
    for (i = 0; i < (order_fullbits/(8*sizeof(digit_t))); ++i) {
        m_cpy[i] = m[i];
    }

    for (i = order_fullbits-order_bits; i > 0; i--) {
        mp_shiftl1(m_cpy, owords);
    }


    for (i = order_bits; i > 0; i--) {
        bit = (unsigned int)(m_cpy[owords-1] >> (RADIX-1));
        mp_shiftl1(m_cpy, owords);
        mask = 0-(digit_t)bit;

        swap_points(xP, xQ, mask);
        xonly_dbl_add(xP, xQ, xR, A24);              // If bit=0 then P <- 2*P and Q <- P+Q, 
        swap_points(xP, xQ, mask);                   // else if bit=1 then Q <- 2*Q and P <- P+Q
    }
}

void compute_a24_mont(f2elm_t a, f2elm_t a24) {
    fp2copy751(a, a24);
    fpadd751(a24[0], two_mont, a24[0]);
    fp2mul751_mont(a24, f2_four_inv_mont, a24);
}

