/********************************************************************************************
*
* Abstract: Internal header file adapted from the Microsoft SIDH library and modified by
*           David Urbanik
*
*********************************************************************************************/  

#ifndef __DEFINITIONS_H__
#define __DEFINITIONS_H__


// For C++
#ifdef __cplusplus
extern "C" {
#endif


#include "system_configuration.h"   
    


typedef digit_t felm_t[NWORDS_FIELD];                             // Datatype for representing 751-bit field elements (768-bit max.)
typedef digit_t dfelm_t[2*NWORDS_FIELD];                          // Datatype for representing double-precision 2x751-bit field elements (2x768-bit max.) 
typedef felm_t  f2elm_t[2];                                       // Datatype for representing quadratic extension field elements GF(p751^2)

typedef struct { f2elm_t n; f2elm_t d; } f2elm_proj;
typedef f2elm_proj f2elm_proj_t[1];

typedef struct { f2elm_t x; f2elm_t y; } point_affine;            // Point representation in affine coordinates on Montgomery curve.
typedef point_affine point_t[1]; 

/*        
typedef struct { f2elm_t X; f2elm_t Z; } point_proj;              // Point representation in projective XZ Montgomery coordinates.
typedef point_proj point_proj_t[1];
*/

typedef struct { f2elm_proj_t X; f2elm_proj_t Y; } full_point_proj;
typedef full_point_proj full_point_proj_t[1];
 

#define UNREFERENCED_PARAMETER(PAR) (PAR)    
#define NBITS_TO_NWORDS(nbits)      (((nbits)+(sizeof(digit_t)*8)-1)/(sizeof(digit_t)*8))


/********************** Constant-time unsigned comparisons ***********************/

// The following functions return 1 (TRUE) if condition is true, 0 (FALSE) otherwise

static __inline unsigned int is_digit_nonzero_ct(digit_t x)
{ // Is x != 0?
    return (unsigned int)((x | (0-x)) >> (RADIX-1));
}

static __inline unsigned int is_digit_zero_ct(digit_t x)
{ // Is x = 0?
    return (unsigned int)(1 ^ is_digit_nonzero_ct(x));
}

static __inline unsigned int is_digit_lessthan_ct(digit_t x, digit_t y)
{ // Is x < y?
    return (unsigned int)((x ^ ((x ^ y) | ((x - y) ^ y))) >> (RADIX-1)); 
}


/********************** Macros for platform-dependent operations **********************/

#if defined(GENERIC_IMPLEMENTATION)
//  ***********************************
//  BEGIN GENERIC IMPLEMENTATION MACROS
//  ***********************************

// Digit multiplication
#define MUL(multiplier, multiplicand, hi, lo)                                                     \
    digit_x_digit((multiplier), (multiplicand), &(lo));
    
// Digit addition with carry
#define ADDC(carryIn, addend1, addend2, carryOut, sumOut)                                         \
    { digit_t tempReg = (addend1) + (digit_t)(carryIn);                                           \
    (sumOut) = (addend2) + tempReg;                                                               \
    (carryOut) = (is_digit_lessthan_ct(tempReg, (digit_t)(carryIn)) | is_digit_lessthan_ct((sumOut), tempReg)); }

// Digit subtraction with borrow
#define SUBC(borrowIn, minuend, subtrahend, borrowOut, differenceOut)                             \
    { digit_t tempReg = (minuend) - (subtrahend);                                                 \
    unsigned int borrowReg = (is_digit_lessthan_ct((minuend), (subtrahend)) | ((borrowIn) & is_digit_zero_ct(tempReg)));  \
    (differenceOut) = tempReg - (digit_t)(borrowIn);                                              \
    (borrowOut) = borrowReg; }
    
// Shift right with flexible datatype
#define SHIFTR(highIn, lowIn, shift, shiftOut, DigitSize)                                         \
    (shiftOut) = ((lowIn) >> (shift)) ^ ((highIn) << (DigitSize - (shift)));
    
// Shift left with flexible datatype
#define SHIFTL(highIn, lowIn, shift, shiftOut, DigitSize)                                         \
    (shiftOut) = ((highIn) << (shift)) ^ ((lowIn) >> (DigitSize - (shift)));

// 64x64-bit multiplication
#define MUL128(multiplier, multiplicand, product)                                                 \
    mp_mul((digit_t*)&(multiplier), (digit_t*)&(multiplicand), (digit_t*)&(product), NWORDS_FIELD/2);

// 128-bit addition, inputs < 2^127
#define ADD128(addend1, addend2, addition)                                                        \
    mp_add((digit_t*)(addend1), (digit_t*)(addend2), (digit_t*)(addition), NWORDS_FIELD);

// 128-bit addition with output carry
#define ADC128(addend1, addend2, carry, addition)                                                 \
    (carry) = mp_add((digit_t*)(addend1), (digit_t*)(addend2), (digit_t*)(addition), NWORDS_FIELD);

//  ***********************************
//  END GENERIC IMPLEMENTATION MACROS**
//  ***********************************

#elif (TARGET == TARGET_ARM64)

//  ***********************************
//  BEGIN ARM64 IMPLEMENTATION MACROS**
//  ***********************************


// Digit multiplication
#define MUL(multiplier, multiplicand, hi, lo)                                                     \
    { uint128_t tempReg = (uint128_t)(multiplier) * (uint128_t)(multiplicand);                    \
    *(hi) = (digit_t)(tempReg >> RADIX);                                                          \
    (lo) = (digit_t)tempReg; }

    
// Digit addition with carry
#define ADDC(carryIn, addend1, addend2, carryOut, sumOut)                                         \
    { uint128_t tempReg = (uint128_t)(addend1) + (uint128_t)(addend2) + (uint128_t)(carryIn);     \
    (carryOut) = (unsigned int)(tempReg >> RADIX);                                                     \
    (sumOut) = (digit_t)tempReg; }  

/*

  NOTE: Some versions of GCC will not compile the code correctly with the above version of ADDC. In this case,
  switch the above version with this one.
    
#define ADDC(carryIn, addend1, addend2, carryOut, sumOut)                                         \
    { digit_t tempReg = (addend1) + (digit_t)(carryIn);                                           \
    (sumOut) = (addend2) + tempReg;                                                               \
    (carryOut) = (is_digit_lessthan_ct(tempReg, (digit_t)(carryIn)) | is_digit_lessthan_ct((sumOut), tempReg)); }

*/

// Digit subtraction with borrow

#define SUBC(borrowIn, minuend, subtrahend, borrowOut, differenceOut)                             \
    { uint128_t tempReg = (uint128_t)(minuend) - (uint128_t)(subtrahend) - (uint128_t)(borrowIn); \
    (borrowOut) = (digit_t)(tempReg >> (sizeof(uint128_t)*8 - 1));                                \
    (differenceOut) = (digit_t)tempReg; }

/*

  NOTE: Some versions of GCC will not compile the code correctly with the above version of SUBC. In this case,
  switch the above version with this one.

#define SUBC(borrowIn, minuend, subtrahend, borrowOut, differenceOut)                             \
    { digit_t tempReg = (minuend) - (subtrahend);                                                 \
    unsigned int borrowReg = (is_digit_lessthan_ct((minuend), (subtrahend)) | ((borrowIn) & is_digit_zero_ct(tempReg)));  \
    (differenceOut) = tempReg - (digit_t)(borrowIn);                                              \
    (borrowOut) = borrowReg; }

*/

// Shift right with flexible datatype
#define SHIFTR(highIn, lowIn, shift, shiftOut, DigitSize)                                         \
    (shiftOut) = ((lowIn) >> (shift)) ^ ((highIn) << (DigitSize - (shift)));
    
// Shift left with flexible datatype
#define SHIFTL(highIn, lowIn, shift, shiftOut, DigitSize)                                         \
    (shiftOut) = ((highIn) << (shift)) ^ ((lowIn) >> (DigitSize - (shift)));


//  ***********************************
//  END ARM64 IMPLEMENTATION MACROS****
//  ***********************************

#elif (TARGET == TARGET_AMD64 && OS_TARGET == OS_WIN)

//  *******************************************
//  BEGIN AMD64/Windows IMPLEMENTATION MACROS**
//  *******************************************

// Digit multiplication
#define MUL(multiplier, multiplicand, hi, lo)                                                     \
    (lo) = _umul128((multiplier), (multiplicand), (hi));                

// Digit addition with carry
#define ADDC(carryIn, addend1, addend2, carryOut, sumOut)                                         \
    (carryOut) = _addcarry_u64((carryIn), (addend1), (addend2), &(sumOut));

// Digit subtraction with borrow
#define SUBC(borrowIn, minuend, subtrahend, borrowOut, differenceOut)                             \
    (borrowOut) = _subborrow_u64((borrowIn), (minuend), (subtrahend), &(differenceOut));

// Digit shift right
#define SHIFTR(highIn, lowIn, shift, shiftOut, DigitSize)                                         \
    (shiftOut) = __shiftright128((lowIn), (highIn), (shift));

// Digit shift left
#define SHIFTL(highIn, lowIn, shift, shiftOut, DigitSize)                                         \
    (shiftOut) = __shiftleft128((lowIn), (highIn), (shift));

// 64x64-bit multiplication
#define MUL128(multiplier, multiplicand, product)                                                 \
    (product)[0] = _umul128((multiplier), (multiplicand), &(product)[1]);

// 128-bit addition, inputs < 2^127
#define ADD128(addend1, addend2, addition)                                                        \
    { unsigned char carry = _addcarry_u64(0, (addend1)[0], (addend2)[0], &(addition)[0]);         \
    _addcarry_u64(carry, (addend1)[1], (addend2)[1], &(addition)[1]); }

// 128-bit addition with output carry
#define ADC128(addend1, addend2, carry, addition)                                                 \
    (carry) = _addcarry_u64(0, (addend1)[0], (addend2)[0], &(addition)[0]);                       \
    (carry) = _addcarry_u64((carry), (addend1)[1], (addend2)[1], &(addition)[1]); 

// 128-bit subtraction, subtrahend < 2^127
#define SUB128(minuend, subtrahend, difference)                                                   \
    { unsigned char borrow = _subborrow_u64(0, (minuend)[0], (subtrahend)[0], &(difference)[0]);  \
    _subborrow_u64(borrow, (minuend)[1], (subtrahend)[1], &(difference)[1]); }

// 128-bit right shift, max. shift value is 64
#define SHIFTR128(Input, shift, shiftOut)                                                         \
    (shiftOut)[0]  = __shiftright128((Input)[0], (Input)[1], (shift));                            \
    (shiftOut)[1] = (Input)[1] >> (shift);    

// 128-bit left shift, max. shift value is 64
#define SHIFTL128(Input, shift, shiftOut)                                                         \
    (shiftOut)[1]  = __shiftleft128((Input)[0], (Input)[1], (shift));                             \
    (shiftOut)[0] = (Input)[0] << (shift);  

#define MULADD128(multiplier, multiplicand, addend, carry, result);    \
    { uint128_t product;                                               \
      MUL128(multiplier, multiplicand, product);                       \
      ADC128(addend, product, carry, result); }   

//  *****************************************
//  END AMD64/Windows IMPLEMENTATION MACROS**
//  *****************************************

#elif (TARGET == TARGET_AMD64 && OS_TARGET == OS_LINUX)

//  *****************************************
//  BEGIN AMD64/Linux IMPLEMENTATION MACROS**
//  *****************************************

// Digit multiplication
#define MUL(multiplier, multiplicand, hi, lo)                                                     \
    { uint128_t tempReg = (uint128_t)(multiplier) * (uint128_t)(multiplicand);                    \
    *(hi) = (digit_t)(tempReg >> RADIX);                                                          \
    (lo) = (digit_t)tempReg; }

// Digit addition with carry
#define ADDC(carryIn, addend1, addend2, carryOut, sumOut)                                         \
    { uint128_t tempReg = (uint128_t)(addend1) + (uint128_t)(addend2) + (uint128_t)(carryIn);     \
    (carryOut) = (digit_t)(tempReg >> RADIX);                                                     \
    (sumOut) = (digit_t)tempReg; }  
    
// Digit subtraction with borrow
#define SUBC(borrowIn, minuend, subtrahend, borrowOut, differenceOut)                             \
    { uint128_t tempReg = (uint128_t)(minuend) - (uint128_t)(subtrahend) - (uint128_t)(borrowIn); \
    (borrowOut) = (digit_t)(tempReg >> (sizeof(uint128_t)*8 - 1));                                \
    (differenceOut) = (digit_t)tempReg; }

// Digit shift right
#define SHIFTR(highIn, lowIn, shift, shiftOut, DigitSize)                                         \
    (shiftOut) = ((lowIn) >> (shift)) ^ ((highIn) << (RADIX - (shift)));

// Digit shift left
#define SHIFTL(highIn, lowIn, shift, shiftOut, DigitSize)                                         \
    (shiftOut) = ((highIn) << (shift)) ^ ((lowIn) >> (RADIX - (shift)));

//  *****************************************
//  END AMD64/Linux IMPLEMENTATION MACROS****
//  *****************************************

#endif


// Multiprecision multiplication selection
#if (TARGET == TARGET_AMD64 || TARGET == TARGET_ARM64)
    #define mp_mul_comba         mp_mul
#else
    #define mp_mul_schoolbook    mp_mul
#endif



/**************** Function prototypes ****************/
/************* Multiprecision functions **************/ 

// Copy wordsize digits, c = a, where lng(a) = nwords
void copy_words(digit_t* a, digit_t* c, unsigned int nwords);

// Multiprecision addition, c = a+b, where lng(a) = lng(b) = nwords. Returns the carry bit 
extern unsigned int mp_add(digit_t* a, digit_t* b, digit_t* c, unsigned int nwords);

// Multiprecision subtraction, c = a-b, where lng(a) = lng(b) = nwords. Returns the borrow bit 
extern unsigned int mp_sub(digit_t* a, digit_t* b, digit_t* c, unsigned int nwords);

// Multiprecision right shift by one
void mp_shiftr1(digit_t* x, unsigned int nwords);

// Multiprecision left right shift by one    
void mp_shiftl1(digit_t* x, unsigned int nwords);

// Digit multiplication, digit * digit -> 2-digit result
void digit_x_digit(digit_t a, digit_t b, digit_t* c);    

// Multiprecision schoolbook multiply, c = a*b, where lng(a) = lng(b) = nwords.
void mp_mul_schoolbook(digit_t* a, digit_t* b, digit_t* c, unsigned int nwords);

// Multiprecision comba multiply, c = a*b, where lng(a) = lng(b) = nwords.
void mp_mul_comba(digit_t* a, digit_t* b, digit_t* c, unsigned int nwords);

/************ Field arithmetic functions *************/

void init_field_constants();

// Copy of a field element, c = a
void fpcopy751(felm_t a, felm_t c);
void fp2copy751(f2elm_t a, f2elm_t c);
void fp2copy751_proj(f2elm_proj_t a, f2elm_proj_t c);

// Zeroing a field element, a = 0
void fpzero751(felm_t a);

// Modular addition, c = a+b mod p751
extern void fpadd751(digit_t* a, digit_t* b, digit_t* c);
extern void fpadd751_asm(digit_t* a, digit_t* b, digit_t* c);
extern void fpadd751_arm64_asm(digit_t* a, digit_t* b, digit_t* c);

// 64-bit arm assembly multiplication
extern void fpmul751_arm64_asm(digit_t* a, digit_t* b, digit_t* c);

// Modular subtraction, c = a-b mod p751
extern void fpsub751(digit_t* a, digit_t* b, digit_t* c);
extern void fpsub751_asm(digit_t* a, digit_t* b, digit_t* c);
extern void fpsub751_arm64_asm(digit_t* a, digit_t* b, digit_t* c);

// Modular negation, a = -a mod p751        
extern void fpneg751(digit_t* a);  

// Modular division by two, c = a/2 mod p751.
void fpdiv2_751(digit_t* a, digit_t* c);

// 751-bit Montgomery reduction, c = a mod p
void rdc_mont(digit_t* a, digit_t* c);
            
// Field multiplication using Montgomery arithmetic, c = a*b*R^-1 mod p751, where R=2^768
void fpmul751_mont(felm_t a, felm_t b, felm_t c);
void mul751_asm(felm_t a, felm_t b, dfelm_t c);
void rdc751_asm(dfelm_t ma, dfelm_t mc);
void rdc751_arm64_asm(dfelm_t ma, dfelm_t mc);
   
// Field squaring using Montgomery arithmetic, c = a*b*R^-1 mod p751, where R=2^768
void fpsqr751_mont(felm_t ma, felm_t mc);

// Conversion to Montgomery representation
void to_mont(felm_t a, felm_t mc);
    
// Conversion from Montgomery representation to standard representation
void from_mont(felm_t ma, felm_t c);

// Field inversion, a = a^-1 in GF(p751)
void fpinv751_mont(felm_t a);

/************ GF(p^2) arithmetic functions *************/
    
// Copy of a GF(p751^2) element, c = a
void fp2copy751(f2elm_t a, f2elm_t c);

// Zeroing a GF(p751^2) element, a = 0
void fp2zero751(f2elm_t a);

// GF(p751^2) negation, a = -a in GF(p751^2)
void fp2neg751(f2elm_t a);

// GF(p751^2) addition, c = a+b in GF(p751^2)
extern void fp2add751(f2elm_t a, f2elm_t b, f2elm_t c);           

// GF(p751^2) subtraction, c = a-b in GF(p751^2)
extern void fp2sub751(f2elm_t a, f2elm_t b, f2elm_t c); 

// GF(p751^2) division by two, c = a/2  in GF(p751^2) 
void fp2div2_751(f2elm_t a, f2elm_t c);
            
// GF(p751^2) squaring using Montgomery arithmetic, c = a^2 in GF(p751^2)
void fp2sqr751_mont(f2elm_t a, f2elm_t c);
 
// GF(p751^2) multiplication using Montgomery arithmetic, c = a*b in GF(p751^2)
void fp2mul751_mont(f2elm_t a, f2elm_t b, f2elm_t c);

// GF(p751^2) exponentiation
void fp2exp751_mont(f2elm_t base, dfelm_t e, int exp_len, f2elm_t result);
    
//  GF(p751^2) square root
void try_fp2sqrt751_mont(f2elm_t a, f2elm_t result, int *is_residue);

// Conversion of a GF(p751^2) element to Montgomery representation
void to_fp2mont(f2elm_t a, f2elm_t mc);

// Conversion of a GF(p751^2) element from Montgomery representation to standard representation
void from_fp2mont(f2elm_t ma, f2elm_t c);

// GF(p751^2) inversion using Montgomery arithmetic, a = (a0-i*a1)/(a0^2+a1^2)
void fp2inv751_mont(f2elm_t a);

// Select either x or y depending on value of option 
void select_f2elm(f2elm_t x, f2elm_t y, f2elm_t z, digit_t option);

unsigned int is_felm_eq(felm_t x, felm_t y);
unsigned int is_f2elm_eq(f2elm_t x, f2elm_t y);
unsigned int is_f2elm_proj_eq(f2elm_proj_t x, f2elm_proj_t y);
unsigned int is_frac_eq(f2elm_t n1, f2elm_t d1, f2elm_t n2, f2elm_t d2);


void fp2projadd751_mont(f2elm_proj_t a, f2elm_proj_t b, f2elm_proj_t c);
void fp2projmul751_mont(f2elm_proj_t a, f2elm_proj_t b, f2elm_proj_t c);
void fp2projsqr751_mont(f2elm_proj_t a, f2elm_proj_t c);
void fp2projsub751_mont(f2elm_proj_t a, f2elm_proj_t b, f2elm_proj_t c);
void fp2projinv751_mont(f2elm_proj_t a);

void fp2mixedadd751_mont(f2elm_t a, f2elm_proj_t b, f2elm_proj_t c);
void fp2mixedsub751_mont_1(f2elm_t a, f2elm_proj_t b, f2elm_proj_t c);
void fp2mixedsub751_mont_2(f2elm_proj_t a, f2elm_t b, f2elm_proj_t c);
void fp2mixedmul751_mont(f2elm_t a, f2elm_proj_t b, f2elm_proj_t c);

void swap_points(f2elm_proj_t xP, f2elm_proj_t xQ, digit_t option);

/************ Elliptic functions *************/

void copy_point(full_point_proj_t P, full_point_proj_t Q);

// Simultaneous doubling and differential addition.
void xonly_dbl_add(f2elm_proj_t xP, f2elm_proj_t xQ, f2elm_proj_t xPQ, f2elm_t A24);

// Doubling of a Montgomery point in projective coordinates (n:d).
void xonly_dbl(f2elm_proj_t xP, f2elm_proj_t xQ, f2elm_t A24, f2elm_t C24);

// Computes [2^e](n:d) on Montgomery curve with projective constant via e repeated doublings.
void xonly_dbl_e(f2elm_proj_t xP, f2elm_proj_t xQ, f2elm_t A, f2elm_t C, int e);

// Tripling of a Montgomery point in projective coordinates (n:d).
void xonly_tpl(f2elm_proj_t xP, f2elm_proj_t xQ, f2elm_t A, f2elm_t C);

// Computes [3^e](n:d) on Montgomery curve with projective constant via e repeated triplings.
void xonly_tpl_e(f2elm_proj_t xP, f2elm_proj_t xQ, f2elm_t A, f2elm_t C, int e);

void ladder(f2elm_proj_t xR, digit_t *m, f2elm_proj_t xP, f2elm_proj_t xQ, f2elm_t A24, unsigned int order_bits, unsigned int order_fullbits);

void get_y2_mont(f2elm_proj_t x1, f2elm_proj_t x2, f2elm_proj_t x3, f2elm_proj_t y1, f2elm_t a, f2elm_t b_inv, f2elm_proj_t y2);

void add_points_generic(f2elm_t a, f2elm_t b_inv, full_point_proj_t P, full_point_proj_t Q, full_point_proj_t R);

void compute_a24_mont(f2elm_t a, f2elm_t a24);

void get_generators_three_tor(f2elm_t a, f2elm_t a24, f2elm_t b_inv, full_point_proj_t Q1, full_point_proj_t Q2);

void get_generators_two_tor(f2elm_t a, f2elm_t a24, f2elm_t b_inv, full_point_proj_t Q1, full_point_proj_t Q2);

void get_y_coord_mont(f2elm_t x, f2elm_t a, f2elm_t b_inv, f2elm_t y, int *on_curve);

void weil_pairing_two_tor(f2elm_t a, f2elm_t b_inv, full_point_proj_t P, full_point_proj_t Q, f2elm_t result);

void weil_pairing_three_tor(f2elm_t a, f2elm_t b_inv, full_point_proj_t P, full_point_proj_t Q, f2elm_t result);

void eval_f_miller(digit_t *subgp_order, int ord_len, f2elm_t a, f2elm_t b_inv, full_point_proj_t P, full_point_proj_t S, f2elm_proj_t result);

void multi_eval_f_miller(digit_t *subgp_order, int ord_len, f2elm_t a, f2elm_t b_inv, full_point_proj_t P, full_point_proj_t *S, f2elm_proj_t *result, int eval_num);

void get_affine(f2elm_proj_t f_proj, f2elm_t f_affine);

#ifdef __cplusplus
}
#endif


#endif
