/********************************************************************************************
*
* Abstract: core functions over GF(p751^2) and field operations over the prime p751
*
* Main code adapted from the SIDH library written by Microsoft, but some code, notably the
* exponentiation function, the sqrt function, and the projective and mixed field arithmetic
* functions, written by David Urbanik (dburbani@uwaterloo.ca).
*
*********************************************************************************************/ 

#include "definitions.h"
#include "constants.h"    


/*******************************************************/
/************* Field arithmetic functions **************/

void init_field_constants() {
    int i;

    //  set felm_t zero
    for (i = 0; i < NWORDS_FIELD; i++) {
        zero[i] = 0;
    }

    //  set f2elm_t f2_zero
    for (i = 0; i < NWORDS_FIELD; i++) {
        f2_zero[0][i] = 0;
    }
    for (i = 0; i < NWORDS_FIELD; i++) {
        f2_zero[1][i] = 0;
    }

    //  set f2elm_t f2_one_mont
    fpcopy751(one_mont, f2_one_mont[0]);
    fpcopy751(zero, f2_one_mont[1]);
}

__inline void fpcopy751(felm_t a, felm_t c)
{ // Copy of a field element, c = a
    unsigned int i;

    for (i = 0; i < NWORDS_FIELD; i++)
        c[i] = a[i];
}

__inline void fpzero751(felm_t a)
{ // Zeroing a field element, a = 0
    unsigned int i;

    for (i = 0; i < NWORDS_FIELD; i++)
        a[i] = 0;
}


void to_mont(felm_t a, felm_t mc)
{ // Conversion to Montgomery representation
  // mc = a*R^2*R^-1 mod p751 = a*R mod p751, where a in [0, p751-1]
  // The Montgomery constant R^2 mod p751 is the global value "Montgomery_R2". 

    fpmul751_mont(a, (digit_t*)&Montgomery_R2, mc);
}


void from_mont(felm_t ma, felm_t c)
{ // Conversion from Montgomery representation to standard representation
  // c = ma*R^-1 mod p751 = a mod p751, where ma in [0, p751-1].
    digit_t one[NWORDS_FIELD] = {0};
    
    one[0] = 1;
    fpmul751_mont(ma, one, c);
}


__inline unsigned int is_felm_eq(felm_t x, felm_t y) {
    //  is x == y?  Does not run in constant time.
    int i;

    for (i = 0; i < NWORDS_FIELD; ++i) {
        if (x[i] != y[i]) return false;
    }
    return true;
}

__inline unsigned int is_f2elm_eq(f2elm_t x, f2elm_t y) {
    return (is_felm_eq(x[0], y[0]) && is_felm_eq(x[1], y[1]));
}

unsigned int is_f2elm_proj_eq(f2elm_proj_t x, f2elm_proj_t y) {
    f2elm_t tmp1, tmp2;

    fp2mul751_mont(x->n, y->d, tmp1);
    fp2mul751_mont(y->n, x->d, tmp2);

    return is_f2elm_eq(tmp1, tmp2);
}

unsigned int is_frac_eq(f2elm_t n1, f2elm_t d1, f2elm_t n2, f2elm_t d2) {
    f2elm_t tmp1, tmp2;

    fp2mul751_mont(n1, d2, tmp1);
    fp2mul751_mont(n2, d1, tmp2);

    return is_f2elm_eq(tmp1, tmp2);
}


void copy_words(digit_t* a, digit_t* c, unsigned int nwords)
{ // Copy wordsize digits, c = a, where lng(a) = nwords
    unsigned int i;
        
    for (i = 0; i < nwords; i++) {                      
        c[i] = a[i];
    }
}


__inline unsigned int mp_sub(digit_t* a, digit_t* b, digit_t* c, unsigned int nwords)
{ // Multiprecision subtraction, c = a-b, where lng(a) = lng(b) = nwords. Returns the borrow bit 
    unsigned int i, borrow = 0;

    for (i = 0; i < nwords; i++) {
        SUBC(borrow, a[i], b[i], borrow, c[i]);
    }

    return borrow;
}


__inline unsigned int mp_add(digit_t* a, digit_t* b, digit_t* c, unsigned int nwords)
{ // Multiprecision addition, c = a+b, where lng(a) = lng(b) = nwords. Returns the carry bit 
    unsigned int i, carry = 0;
        
    for (i = 0; i < nwords; i++) {                      
        ADDC(carry, a[i], b[i], carry, c[i]);
    }

    return carry;
}


void mp_shiftr1(digit_t* x, unsigned int nwords)
{ // Multiprecision right shift by one
    unsigned int i;

    for (i = 0; i < nwords-1; i++) {
        SHIFTR(x[i+1], x[i], 1, x[i], RADIX);
    }
    x[nwords-1] >>= 1;
}


void mp_shiftl1(digit_t* x, unsigned int nwords)
{ // Multiprecision left right shift by one
    int i;

    for (i = nwords-1; i > 0; i--) {
        SHIFTL(x[i], x[i-1], 1, x[i], RADIX);
    }
    x[0] <<= 1;
}


void fpmul751_mont(felm_t ma, felm_t mb, felm_t mc)
{ // 751-bit Comba multi-precision multiplication, c = a*b mod p751
    dfelm_t temp = {0};

    mp_mul(ma, mb, temp, NWORDS_FIELD);
    rdc_mont(temp, mc);
}


void fpsqr751_mont(felm_t ma, felm_t mc)
{ // 751-bit Comba multi-precision squaring, c = a^2 mod p751
    dfelm_t temp = {0};

    mp_mul(ma, ma, temp, NWORDS_FIELD);
    rdc_mont(temp, mc);
}


void fpinv751_mont(felm_t a)
{// Field inversion using Montgomery arithmetic, a = a^-1*R mod p751
    felm_t t[27], tt;
    unsigned int i, j;
    
    // Precomputed table
    fpsqr751_mont(a, tt);
    fpmul751_mont(a, tt, t[0]);
    fpmul751_mont(t[0], tt, t[1]);
    fpmul751_mont(t[1], tt, t[2]);
    fpmul751_mont(t[2], tt, t[3]); 
    fpmul751_mont(t[3], tt, t[3]);
    for (i = 3; i <= 8; i++) fpmul751_mont(t[i], tt, t[i+1]);
    fpmul751_mont(t[9], tt, t[9]);
    for (i = 9; i <= 20; i++) fpmul751_mont(t[i], tt, t[i+1]);
    fpmul751_mont(t[21], tt, t[21]); 
    for (i = 21; i <= 24; i++) fpmul751_mont(t[i], tt, t[i+1]); 
    fpmul751_mont(t[25], tt, t[25]);
    fpmul751_mont(t[25], tt, t[26]);

    fpcopy751(a, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[20], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[24], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[11], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[8], tt, tt);
    for (i = 0; i < 8; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[2], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[23], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[2], tt, tt);
    for (i = 0; i < 9; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[2], tt, tt);
    for (i = 0; i < 10; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[15], tt, tt);
    for (i = 0; i < 8; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[13], tt, tt);
    for (i = 0; i < 8; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[26], tt, tt);
    for (i = 0; i < 8; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[20], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[11], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[10], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[14], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[4], tt, tt);
    for (i = 0; i < 10; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[18], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[1], tt, tt);
    for (i = 0; i < 7; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[22], tt, tt);
    for (i = 0; i < 10; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[6], tt, tt);
    for (i = 0; i < 7; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[24], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[9], tt, tt);
    for (i = 0; i < 8; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[18], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[17], tt, tt);
    for (i = 0; i < 8; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(a, tt, tt);
    for (i = 0; i < 10; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[16], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[7], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[0], tt, tt);
    for (i = 0; i < 7; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[12], tt, tt);
    for (i = 0; i < 7; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[19], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[22], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[25], tt, tt);
    for (i = 0; i < 7; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[2], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[10], tt, tt);
    for (i = 0; i < 7; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[22], tt, tt);
    for (i = 0; i < 8; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[18], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[4], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[14], tt, tt);
    for (i = 0; i < 7; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[13], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[5], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[23], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[21], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[2], tt, tt);
    for (i = 0; i < 7; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[23], tt, tt);
    for (i = 0; i < 8; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[12], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[9], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[3], tt, tt);
    for (i = 0; i < 7; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[13], tt, tt);
    for (i = 0; i < 7; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[17], tt, tt);
    for (i = 0; i < 8; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[26], tt, tt);
    for (i = 0; i < 8; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[5], tt, tt);
    for (i = 0; i < 8; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[8], tt, tt);
    for (i = 0; i < 8; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[11], tt, tt);
    for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    fpmul751_mont(t[22], tt, tt);
    for (i = 0; i < 7; i++) fpsqr751_mont(tt, tt);
    for (j = 0; j < 61; j++) {
        fpmul751_mont(t[26], tt, tt);
        for (i = 0; i < 6; i++) fpsqr751_mont(tt, tt);
    }
    fpmul751_mont(t[25], tt, a);  
}


/***********************************************/
/************* GF(p^2) FUNCTIONS ***************/

void fp2copy751(f2elm_t a, f2elm_t c)
{// Copy of a GF(p751^2) element, c = a
    fpcopy751(a[0], c[0]);
    fpcopy751(a[1], c[1]);
}

void fp2copy751_proj(f2elm_proj_t a, f2elm_proj_t c) {
    fp2copy751(a->n, c->n);
    fp2copy751(a->d, c->d);
}

void fp2zero751(f2elm_t a)
{// Zeroing a GF(p751^2) element, a = 0
    fpzero751(a[0]);
    fpzero751(a[1]);
}


void fp2neg751(f2elm_t a)
{// GF(p751^2) negation, a = -a in GF(p751^2)
    fpneg751(a[0]);
    fpneg751(a[1]);
}


__inline void fp2add751(f2elm_t a, f2elm_t b, f2elm_t c)           
{// GF(p751^2) addition, c = a+b in GF(p751^2)
    fpadd751(a[0], b[0], c[0]);
    fpadd751(a[1], b[1], c[1]);
}


__inline void fp2sub751(f2elm_t a, f2elm_t b, f2elm_t c)          
{// GF(p751^2) subtraction, c = a-b in GF(p751^2) 
    fpsub751(a[0], b[0], c[0]);
    fpsub751(a[1], b[1], c[1]);
}


void fp2div2_751(f2elm_t a, f2elm_t c)          
{// GF(p751^2) division by two, c = a/2  in GF(p751^2) 
    fpdiv2_751(a[0], c[0]);
    fpdiv2_751(a[1], c[1]);
}


void fp2sqr751_mont(f2elm_t a, f2elm_t c)
{// GF(p751^2) squaring using Montgomery arithmetic, c = a^2 in GF(p751^2)
    felm_t t1, t2, t3;

    mp_add(a[0], a[1], t1, NWORDS_FIELD);    // t1 = a0+a1 
    fpsub751(a[0], a[1], t2);                // t2 = a0-a1
    mp_add(a[0], a[0], t3, NWORDS_FIELD);    // t3 = 2a0
    fpmul751_mont(t1, t2, c[0]);             // c0 = (a0+a1)(a0-a1)
    fpmul751_mont(t3, a[1], c[1]);           // c1 = 2a0*a1
}


void fp2mul751_mont(f2elm_t a, f2elm_t b, f2elm_t c)
{// GF(p751^2) multiplication using Montgomery arithmetic, c = a*b in GF(p751^2)
    felm_t t1, t2;
    dfelm_t tt1, tt2, tt3; 
    digit_t mask;
    unsigned int i, borrow;

    mp_mul(a[0], b[0], tt1, NWORDS_FIELD);           // tt1 = a0*b0
    mp_mul(a[1], b[1], tt2, NWORDS_FIELD);           // tt2 = a1*b1
    mp_add(a[0], a[1], t1, NWORDS_FIELD);            // t1 = a0+a1
    mp_add(b[0], b[1], t2, NWORDS_FIELD);            // t2 = b0+b1
    borrow = mp_sub(tt1, tt2, tt3, 2*NWORDS_FIELD);  // tt3 = a0*b0 - a1*b1
    mask = 0 - (digit_t)borrow;                      // if tt3 < 0 then mask = 0xFF..F, else if tt3 >= 0 then mask = 0x00..0
    borrow = 0;
    for (i = 0; i < NWORDS_FIELD; i++) {             // tt3 = tt3 + (mask & 2^768*p751)
        ADDC(borrow, tt3[NWORDS_FIELD+i], ((digit_t*)p751)[i] & mask, borrow, tt3[NWORDS_FIELD+i]);
    }
    rdc_mont(tt3, c[0]);                             // c[0] = a0*b0 - a1*b1
    mp_add(tt1, tt2, tt1, 2*NWORDS_FIELD);           // tt1 = a0*b0 + a1*b1
    mp_mul(t1, t2, tt2, NWORDS_FIELD);               // tt2 = (a0+a1)*(b0+b1)
    mp_sub(tt2, tt1, tt2, 2*NWORDS_FIELD);           // tt2 = (a0+a1)*(b0+b1) - a0*b0 - a1*b1 
    rdc_mont(tt2, c[1]);                             // c[1] = (a0+a1)*(b0+b1) - a0*b0 - a1*b1 
}

void fp2inv751_mont(f2elm_t a)
{// GF(p751^2) inversion using Montgomery arithmetic, a = (a0-i*a1)/(a0^2+a1^2)
    f2elm_t t1;

    fpsqr751_mont(a[0], t1[0]);             // t10 = a0^2
    fpsqr751_mont(a[1], t1[1]);             // t11 = a1^2
    fpadd751(t1[0], t1[1], t1[0]);          // t10 = a0^2+a1^2
    fpinv751_mont(t1[0]);                   // t10 = (a0^2+a1^2)^-1
    fpneg751(a[1]);                         // a = a0-i*a1
    fpmul751_mont(a[0], t1[0], a[0]);
    fpmul751_mont(a[1], t1[0], a[1]);       // a = (a0-i*a1)*(a0^2+a1^2)^-1
}


void swap_points(f2elm_proj_t xP, f2elm_proj_t xQ, digit_t option)
{ // Swap points
  // If option = 0 then P <- P and Q <- Q, else if option = 0xFF...FF then P <- Q and Q <- P
    digit_t temp;
    unsigned int i;

    for (i = 0; i < NWORDS_FIELD; i++) {
        temp = option & (xP->n[0][i] ^ xQ->n[0][i]);
        xP->n[0][i] = temp ^ xP->n[0][i]; 
        xQ->n[0][i] = temp ^ xQ->n[0][i]; 
        temp = option & (xP->d[0][i] ^ xQ->d[0][i]);
        xP->d[0][i] = temp ^ xP->d[0][i]; 
        xQ->d[0][i] = temp ^ xQ->d[0][i]; 
        temp = option & (xP->n[1][i] ^ xQ->n[1][i]);
        xP->n[1][i] = temp ^ xP->n[1][i]; 
        xQ->n[1][i] = temp ^ xQ->n[1][i]; 
        temp = option & (xP->d[1][i] ^ xQ->d[1][i]);
        xP->d[1][i] = temp ^ xP->d[1][i]; 
        xQ->d[1][i] = temp ^ xQ->d[1][i];
    }
}


void select_f2elm(f2elm_t x, f2elm_t y, f2elm_t z, digit_t option)
{ // Select either x or y depending on value of option 
  // If option = 0 then z <- x, else if option = 0xFF...FF then z <- y
    unsigned int i;

    for (i = 0; i < NWORDS_FIELD; i++) {
        z[0][i] = (option & (x[0][i] ^ y[0][i])) ^ x[0][i]; 
        z[1][i] = (option & (x[1][i] ^ y[1][i])) ^ x[1][i]; 
    }
}

void to_fp2mont(f2elm_t a, f2elm_t mc)
{ // Conversion of a GF(p751^2) element to Montgomery representation
  // mc_i = a_i*R^2*R^-1 = a_i*R in GF(p751^2). 

    to_mont(a[0], mc[0]);
    to_mont(a[1], mc[1]);
}


void from_fp2mont(f2elm_t ma, f2elm_t c)
{ // Conversion of a GF(p751^2) element from Montgomery representation to standard representation
  // c_i = ma_i*R^-1 = a_i in GF(p751^2).

    from_mont(ma[0], c[0]);
    from_mont(ma[1], c[1]);
}

/***********************************************************************************************
*   The following code was written by David Urbanik (dburbani@uwaterloo.ca)
************************************************************************************************/

void fp2projadd751_mont(f2elm_proj_t a, f2elm_proj_t b, f2elm_proj_t c) {
    f2elm_t tmp1, tmp2;

    fp2mul751_mont(a->n, b->d, tmp1);
    fp2mul751_mont(a->d, b->n, tmp2);
    fp2add751(tmp1, tmp2, c->n);
    fp2mul751_mont(a->d, b->d, c->d);
}

void fp2projsqr751_mont(f2elm_proj_t a, f2elm_proj_t c) {
    fp2sqr751_mont(a->n, c->n);
    fp2sqr751_mont(a->d, c->d);
}

void fp2projinv751_mont(f2elm_proj_t a) {
    f2elm_t tmp1;
    fp2copy751(a->d, tmp1);
    fp2copy751(a->n, a->d);
    fp2copy751(tmp1, a->n);
}

void fp2projmul751_mont(f2elm_proj_t a, f2elm_proj_t b, f2elm_proj_t c) {
    fp2mul751_mont(a->n, b->n, c->n);
    fp2mul751_mont(a->d, b->d, c->d);
}

void fp2projsub751_mont(f2elm_proj_t a, f2elm_proj_t b, f2elm_proj_t c) {
    f2elm_t tmp1, tmp2;

    fp2mul751_mont(a->n, b->d, tmp1);
    fp2mul751_mont(a->d, b->n, tmp2);
    fp2sub751(tmp1, tmp2, c->n);
    fp2mul751_mont(a->d, b->d, c->d);
}

void fp2mixedadd751_mont(f2elm_t a, f2elm_proj_t b, f2elm_proj_t c) {
    f2elm_t tmp1;

    fp2mul751_mont(a, b->d, tmp1);
    fp2add751(tmp1, b->n, c->n);
    fp2copy751(b->d, c->d);
}

void fp2mixedsub751_mont_1(f2elm_t a, f2elm_proj_t b, f2elm_proj_t c) {
    f2elm_t tmp1;

    fp2mul751_mont(a, b->d, tmp1);
    fp2sub751(tmp1, b->n, c->n);
    fp2copy751(b->d, c->d);
}

void fp2mixedsub751_mont_2(f2elm_proj_t a, f2elm_t b, f2elm_proj_t c) {
    f2elm_t tmp1;

    fp2mul751_mont(a->d, b, tmp1);
    fp2sub751(a->n, tmp1, c->n);
    fp2copy751(a->d, c->d);
}

void fp2mixedmul751_mont(f2elm_t a, f2elm_proj_t b, f2elm_proj_t c) {
    fp2mul751_mont(a, b->n, c->n);
    fp2copy751(b->d, c->d);
}

void fp2exp751_mont(f2elm_t base, digit_t *e, int exp_len, f2elm_t result) {
    int i, j;
    digit_t mask;

    f2elm_t tmp_result;

    fpcopy751(one_mont, tmp_result[0]);
    fpcopy751(zero, tmp_result[1]);

    for (i = exp_len - 1; i >= 0; i--) {
        mask = 1; 
        mask <<= (sizeof(digit_t)*8 - 1);

        for (j = 0; j < sizeof(digit_t)*8; ++j) {

            fp2sqr751_mont(tmp_result, tmp_result);

            if (mask & e[i]) {
                fp2mul751_mont(tmp_result, base, tmp_result);
            }
            mask >>= 1;
        }
    }

    fpcopy751(tmp_result[0], result[0]);
    fpcopy751(tmp_result[1], result[1]);
}

void fpsqrt751_mont(felm_t a, felm_t result) {
    int i;
    felm_t tmp;
    fpcopy751(a, result);

    for (i = 0; i < 370; i++) {
        fpsqr751_mont(result, result);
    }

    for (i = 0; i < 239; i++) {
        fpsqr751_mont(result, tmp);
        fpmul751_mont(tmp, result, result);
    }
}

//  computes the square root with probability 1/2
void try_fp2sqrt751_mont(f2elm_t a, f2elm_t result, int *is_residue) {
    felm_t tmp1, tmp2, tmp3;
    felm_t b0_sqr_1, b0_sqr_2;
    f2elm_t test;

    fpsqr751_mont(a[0], tmp1);
    fpsqr751_mont(a[1], tmp2);
    fpadd751(tmp1, tmp2, tmp1);
    fpsqrt751_mont(tmp1, tmp2);

    //  check to see if sqrt suceeded -- if it didn't, then a is not a square
    //  and we can return

    fpsqr751_mont(tmp2, tmp3);
    if (!is_felm_eq(tmp3, tmp1)) {
        *is_residue = 0;
        return;
    }


    //  try b0_sqr_1
    fpadd751(a[0], tmp2, b0_sqr_1);
    fpmul751_mont(b0_sqr_1, two_inv_mont, b0_sqr_1);

    fpsqrt751_mont(b0_sqr_1, result[0]);
    fpcopy751(result[0], tmp1);
    fpinv751_mont(tmp1);
    fpmul751_mont(a[1], tmp1, result[1]);
    fpmul751_mont(result[1], two_inv_mont, result[1]);


    fp2sqr751_mont(result, test);

    if (is_f2elm_eq(test, a)) {
        *is_residue = 1;
        return;
    }

    //  try b0_sqr_2
    fpsub751(a[0], tmp2, b0_sqr_2);
    fpmul751_mont(b0_sqr_2, two_inv_mont, b0_sqr_2);

    fpsqrt751_mont(b0_sqr_2, result[0]);
    fpcopy751(result[0], tmp1);
    fpinv751_mont(tmp1);
    fpmul751_mont(a[1], tmp1, result[1]);
    fpmul751_mont(result[1], two_inv_mont, result[1]);

    fp2sqr751_mont(result, test);

    if (is_f2elm_eq(test, a)) {
        *is_residue = 1;
        return;
    }

    *is_residue = 0;
}
