/***************************************************************************************************

Main code for the key compression/decompression library.

Written by David Urbanik (dburbani@uwaterloo.ca)

***************************************************************************************************/

#include "definitions.h"
#include "constants.h"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>


typedef struct { 
    f2elm_t a; 
    digit_t a1[6]; 
    digit_t a2[6];
    digit_t b1[6];
    digit_t b2[6];
} compressed_key;

void print_f2elm(f2elm_t e) {
    int i;

    printf("{");
    for (i = 0; i < NWORDS_FIELD; i++) {
        printf("%lx ", e[0][i]);
    }
    printf("}\n");
    printf("{");
    for (i = 0; i < NWORDS_FIELD; i++) {
        printf("%lx ", e[1][i]);
    }
    printf("}i\n\n");
}

void print_f2elm_mont(f2elm_t e) {
    f2elm_t tmp;
    from_fp2mont(e, tmp);
    print_f2elm(tmp);
}

int count_leading_zero_bits(digit_t *big_num, int len) {
    int i, j;
    int leading_zero_num = 0;

    for (i = len-1; i >= 0; i--) {
        digit_t test_bit = (((digit_t) 0x1) << (sizeof(digit_t)*8-1));

        for (j = 0; j < sizeof(digit_t)*8; j++) {
            if ((big_num[i] & test_bit) == 0) {
                leading_zero_num++;
                test_bit >>= 1;
            } else {
                return leading_zero_num;
            }
        }
    }

    return 0;
}

void comp_lin_comb(f2elm_t a, f2elm_t b_inv, f2elm_t a24, full_point_proj_t P1, full_point_proj_t P2, digit_t *r, digit_t *s, f2elm_proj_t result) {
    int i;
    full_point_proj_t rP1, sP2, sum_point;
    f2elm_proj_t xrp1P1, xsp1P2;
    int r_nonzero = 0;
    int s_nonzero = 0;

    for (i = 0; i < 6; i++) {
        if (r[i] != 0) {
            r_nonzero = 1;
            break;
        }
    }

    for (i = 0; i < 6; i++) {
        if (s[i] != 0) {
            s_nonzero = 1;
            break;
        }
    }

    if (r_nonzero) {
        int leading_zero_bits_r = count_leading_zero_bits(r, 6);
        ladder(P1->X, r, rP1->X, xrp1P1, a24, 384 - leading_zero_bits_r, 384);
    }

    if (s_nonzero) {
        int leading_zero_bits_s = count_leading_zero_bits(s, 6);
        ladder(P2->X, s, sP2->X, xsp1P2, a24, 384 - leading_zero_bits_s, 384);
    }

    if (r_nonzero && s_nonzero) {
        get_y2_mont(P1->X, rP1->X, xrp1P1, P1->Y, a, b_inv, rP1->Y);
        get_y2_mont(P2->X, sP2->X, xsp1P2, P2->Y, a, b_inv, sP2->Y);
        add_points_generic(a, b_inv, rP1, sP2, sum_point);
        fp2copy751_proj(sum_point->X, result);
    } else if (r_nonzero) {
        fp2copy751_proj(rP1->X, result);
    } else if (s_nonzero) {
        fp2copy751_proj(sP2->X, result);
    } else {
        fp2copy751(f2_zero, result->n);
        fp2copy751(f2_one_mont, result->d);
    }
}

void base_3_to_binary(int *a, int a_len, digit_t *result, int result_len) {
    int i, j;
    
    for (i = 0; i < result_len; i++) {
        result[i] = 0;
    }

    result[0] = a[a_len-1];
    for (i = a_len-2; i >= 0; i--) {
        uint128_t tmp;
        unsigned int carry = 0;

        for (j = 0; j < result_len; j++) {
            tmp = 3*((uint128_t) result[j]) + carry;
            result[j] = (digit_t) tmp;
            tmp >>= 64;
            carry = (unsigned int) tmp;
        }

        carry = a[i];
        for (j = 0; j < result_len; j++) {
            tmp = ((uint128_t) result[j]) + carry;
            result[j] = (digit_t) tmp;
            tmp >>= 64;
            carry = (unsigned int) tmp;
        }
    }
}

void discrete_log_three_tor(f2elm_t *b_exp, f2elm_t *b_exp_sqr, f2elm_t p, digit_t *result, int do_inv) {
    int i, j;
    f2elm_t p_exp[239];
    int a[239];
    f2elm_t tmp, tmp2;
    int limb_size = sizeof(digit_t)*8;

    fp2copy751(p, p_exp[0]);
    for (i = 1; i < 239; i++) {
        fp2sqr751_mont(p_exp[i-1], p_exp[i]);
        fp2mul751_mont(p_exp[i-1], p_exp[i], p_exp[i]);
    }

    for (i = 0; i < 239; i++) {
        fp2copy751(f2_one_mont, tmp);

        for (j = 0; j < i; j++) {
            if (a[j] == 1) {
                fp2mul751_mont(tmp, b_exp[j + (238 - i)], tmp);
            } else if (a[j] == 2) {
                fp2mul751_mont(tmp, b_exp_sqr[j + (238 - i)], tmp);
            }
        }

        if (is_f2elm_eq(tmp, p_exp[238-i])) {
            a[i] = 0;
        } else {
            fp2mul751_mont(tmp, b_exp[238], tmp2);
            if (is_f2elm_eq(tmp2, p_exp[238-i])) {
                a[i] = 1;
            } else {
                a[i] = 2;
            }
        }
    }

    if (do_inv) {
        for (i = 0; i < 239; i++) {
            a[i] = (2 - a[i]);
        }

        for (i = 0; i < 239; i++) {
            if (a[i] < 2) {
                a[i]++;
                break;
            } else {
                a[i] = 0;
            }
        }
    }

    base_3_to_binary(a, 239, result, 6);
}

void discrete_log_two_tor(f2elm_t *b_exp, f2elm_t p, digit_t *result, int do_inv) {
    int i, j;
    f2elm_t p_exp[372];
    int a[372];
    f2elm_t tmp;
    int limb_size = sizeof(digit_t)*8;

    fp2copy751(p, p_exp[0]);
    for (i = 1; i < 372; i++) {
        fp2sqr751_mont(p_exp[i-1], p_exp[i]);
    }

    for (i = 0; i < 372; i++) {
        fp2copy751(f2_one_mont, tmp);

        for (j = 0; j < i; j++) {
            if (a[j] == 1) {
                fp2mul751_mont(tmp, b_exp[j + (371 - i)], tmp);
            }
        }

        if (is_f2elm_eq(tmp, p_exp[371-i])) {
            a[i] = 0;
        } else {
            a[i] = 1;
        }
    }
    
    if (do_inv) {
        for (i = 0; i < 372; i++) {
            a[i] = (1 - a[i]);
        }

        for (i = 0; i < 372; i++) {
            if (a[i] == 0) {
                a[i] = 1;
                break;
            } else {
                a[i] = 0;
            }
        }
    }

    int result_len = (372/limb_size) + ((372 % limb_size) != 0);

    for (i = 0; i < result_len; i++) {
        digit_t mask = 1;
        result[i] = 0;
        for (j = 0; j < limb_size && (i*limb_size + j) < 372; ++j) {
            result[i] += a[i*limb_size + j]*mask;
            mask *= 2;
        }
    }
}


//  result array should be length 5
void multi_weil_pairing_two_tor(f2elm_t a, f2elm_t b_inv, full_point_proj_t Q1, full_point_proj_t Q2, full_point_proj_t P1, full_point_proj_t P2, f2elm_t *result) {
    full_point_proj_t argQ1[3];
    full_point_proj_t argQ2[3];
    full_point_proj_t argP1[2];
    full_point_proj_t argP2[2];
    f2elm_proj_t resultQ1[3];
    f2elm_proj_t resultQ2[3];
    f2elm_proj_t resultP1[2];
    f2elm_proj_t resultP2[2];
    f2elm_proj_t tmp;

    copy_point(Q2, argQ1[0]);
    copy_point(P1, argQ1[1]);
    copy_point(P2, argQ1[2]);

    copy_point(Q1, argQ2[0]);
    copy_point(P1, argQ2[1]);
    copy_point(P2, argQ2[2]);

    copy_point(Q1, argP1[0]);
    copy_point(Q2, argP1[1]);

    copy_point(Q1, argP2[0]);
    copy_point(Q2, argP2[1]);

    multi_eval_f_miller(two_exp_372, 372, a, b_inv, Q1, argQ1, resultQ1, 3);
    multi_eval_f_miller(two_exp_372, 372, a, b_inv, Q2, argQ2, resultQ2, 3);
    multi_eval_f_miller(two_exp_372, 372, a, b_inv, P1, argP1, resultP1, 2);
    multi_eval_f_miller(two_exp_372, 372, a, b_inv, P2, argP2, resultP2, 2);

    //  compute e(Q2, Q1) and e(Q1, Q2)
    
    fp2mul751_mont(resultQ2[0]->n, resultQ1[0]->d, tmp->n);
    fp2mul751_mont(resultQ2[0]->d, resultQ1[0]->n, tmp->d);

    fp2inv751_mont(tmp->d);
    fp2mul751_mont(tmp->n, tmp->d, result[0]);

    //  compute e(P1, Q1)

    fp2mul751_mont(resultP1[0]->n, resultQ1[1]->d, tmp->n);
    fp2mul751_mont(resultP1[0]->d, resultQ1[1]->n, tmp->d);
    fp2inv751_mont(tmp->d);
    fp2mul751_mont(tmp->n, tmp->d, result[1]);

    //  compute e(P1, Q2)
    
    fp2mul751_mont(resultP1[1]->n, resultQ2[1]->d, tmp->n);
    fp2mul751_mont(resultP1[1]->d, resultQ2[1]->n, tmp->d);
    fp2inv751_mont(tmp->d);
    fp2mul751_mont(tmp->n, tmp->d, result[2]);

    //  compute e(P2, Q1)

    fp2mul751_mont(resultP2[0]->n, resultQ1[2]->d, tmp->n);
    fp2mul751_mont(resultP2[0]->d, resultQ1[2]->n, tmp->d);
    fp2inv751_mont(tmp->d);
    fp2mul751_mont(tmp->n, tmp->d, result[3]);

    //  compute e(P2, Q2)

    fp2mul751_mont(resultP2[1]->n, resultQ2[2]->d, tmp->n);
    fp2mul751_mont(resultP2[1]->d, resultQ2[2]->n, tmp->d);
    fp2inv751_mont(tmp->d);
    fp2mul751_mont(tmp->n, tmp->d, result[4]);
}

//  result array should be length 5
void multi_weil_pairing_three_tor(f2elm_t a, f2elm_t b_inv, full_point_proj_t Q1, full_point_proj_t Q2, full_point_proj_t P1, full_point_proj_t P2, f2elm_t *result) {
    full_point_proj_t argQ1[3];
    full_point_proj_t argQ2[3];
    full_point_proj_t argP1[2];
    full_point_proj_t argP2[2];
    f2elm_proj_t resultQ1[3];
    f2elm_proj_t resultQ2[3];
    f2elm_proj_t resultP1[2];
    f2elm_proj_t resultP2[2];
    f2elm_proj_t tmp;

    copy_point(Q2, argQ1[0]);
    copy_point(P1, argQ1[1]);
    copy_point(P2, argQ1[2]);

    copy_point(Q1, argQ2[0]);
    copy_point(P1, argQ2[1]);
    copy_point(P2, argQ2[2]);

    copy_point(Q1, argP1[0]);
    copy_point(Q2, argP1[1]);

    copy_point(Q1, argP2[0]);
    copy_point(Q2, argP2[1]);

    multi_eval_f_miller(three_exp_239, 378, a, b_inv, Q1, argQ1, resultQ1, 3);
    multi_eval_f_miller(three_exp_239, 378, a, b_inv, Q2, argQ2, resultQ2, 3);
    multi_eval_f_miller(three_exp_239, 378, a, b_inv, P1, argP1, resultP1, 2);
    multi_eval_f_miller(three_exp_239, 378, a, b_inv, P2, argP2, resultP2, 2);

    //  compute e(Q2, Q1) and e(Q1, Q2)
    
    fp2mul751_mont(resultQ2[0]->n, resultQ1[0]->d, tmp->n);
    fp2mul751_mont(resultQ2[0]->d, resultQ1[0]->n, tmp->d);

    fp2inv751_mont(tmp->d);
    fp2mul751_mont(tmp->n, tmp->d, result[0]);

    fp2neg751(result[0]);

    //  compute e(P1, Q1)

    fp2mul751_mont(resultP1[0]->n, resultQ1[1]->d, tmp->n);
    fp2mul751_mont(resultP1[0]->d, resultQ1[1]->n, tmp->d);
    fp2inv751_mont(tmp->d);
    fp2mul751_mont(tmp->n, tmp->d, result[1]);
    fp2neg751(result[1]);

    //  compute e(P1, Q2)
    
    fp2mul751_mont(resultP1[1]->n, resultQ2[1]->d, tmp->n);
    fp2mul751_mont(resultP1[1]->d, resultQ2[1]->n, tmp->d);
    fp2inv751_mont(tmp->d);
    fp2mul751_mont(tmp->n, tmp->d, result[2]);
    fp2neg751(result[2]);

    //  compute e(P2, Q1)

    fp2mul751_mont(resultP2[0]->n, resultQ1[2]->d, tmp->n);
    fp2mul751_mont(resultP2[0]->d, resultQ1[2]->n, tmp->d);
    fp2inv751_mont(tmp->d);
    fp2mul751_mont(tmp->n, tmp->d, result[3]);
    fp2neg751(result[3]);

    //  compute e(P2, Q2)

    fp2mul751_mont(resultP2[1]->n, resultQ2[2]->d, tmp->n);
    fp2mul751_mont(resultP2[1]->d, resultQ2[2]->n, tmp->d);
    fp2inv751_mont(tmp->d);
    fp2mul751_mont(tmp->n, tmp->d, result[4]);
    fp2neg751(result[4]);
}


void compress_three_tor(f2elm_t a, f2elm_t b_inv, f2elm_t x1, f2elm_t x2, compressed_key *key) {
    int i;
    f2elm_t a24;
    compute_a24_mont(a, a24);

    full_point_proj_t P1, P2;
    full_point_proj_t Q1, Q2;
    fp2copy751(f2_one_mont, P1->X->d);
    fp2copy751(f2_one_mont, P1->Y->d);
    fp2copy751(f2_one_mont, P2->X->d);
    fp2copy751(f2_one_mont, P2->Y->d);

    fp2copy751(a, key->a);
    fp2copy751(x1, P1->X->n);
    fp2copy751(x2, P2->X->n);


    int on_curve;
    get_y_coord_mont(P1->X->n, a, b_inv, P1->Y->n, &on_curve);
    get_y_coord_mont(P2->X->n, a, b_inv, P2->Y->n, &on_curve);


    get_generators_three_tor(a, a24, b_inv, Q1, Q2);

    f2elm_t pairing[5];
    multi_weil_pairing_three_tor(a, b_inv, Q1, Q2, P1, P2, pairing);


    f2elm_t base_exp[239];
    f2elm_t base_exp_sqr[239];

    fp2copy751(pairing[0], base_exp[0]);
    fp2sqr751_mont(base_exp[0], base_exp_sqr[0]);
    for (i = 1; i < 239; i++) {
        fp2mul751_mont(base_exp_sqr[i-1], base_exp[i-1], base_exp[i]);
        fp2sqr751_mont(base_exp[i], base_exp_sqr[i]);
    }

    discrete_log_three_tor(base_exp, base_exp_sqr, pairing[1], key->a2, 0);
    discrete_log_three_tor(base_exp, base_exp_sqr, pairing[2], key->a1, 1);
    discrete_log_three_tor(base_exp, base_exp_sqr, pairing[3], key->b2, 0);
    discrete_log_three_tor(base_exp, base_exp_sqr, pairing[4], key->b1, 1);
}

void compress_two_tor(f2elm_t a, f2elm_t b_inv, f2elm_t x1, f2elm_t x2, compressed_key *key) {
    int i;
    f2elm_t a24;
    compute_a24_mont(a, a24);

    full_point_proj_t P1, P2;
    full_point_proj_t Q1, Q2;
    fp2copy751(f2_one_mont, P1->X->d);
    fp2copy751(f2_one_mont, P1->Y->d);
    fp2copy751(f2_one_mont, P2->X->d);
    fp2copy751(f2_one_mont, P2->Y->d);

    fp2copy751(a, key->a);
    fp2copy751(x1, P1->X->n);
    fp2copy751(x2, P2->X->n);

    int on_curve;
    get_y_coord_mont(P1->X->n, a, b_inv, P1->Y->n, &on_curve);
    get_y_coord_mont(P2->X->n, a, b_inv, P2->Y->n, &on_curve);


    get_generators_two_tor(a, a24, b_inv, Q1, Q2);


    f2elm_t pairing[5];
    multi_weil_pairing_two_tor(a, b_inv, Q1, Q2, P1, P2, pairing);

    f2elm_t base_exp[372];

    fp2copy751(pairing[0], base_exp[0]);
    for (i = 1; i < 372; i++) {
        fp2sqr751_mont(base_exp[i-1], base_exp[i]);
    }

    discrete_log_two_tor(base_exp, pairing[1], key->a2, 0);
    discrete_log_two_tor(base_exp, pairing[2], key->a1, 1);
    discrete_log_two_tor(base_exp, pairing[3], key->b2, 0);
    discrete_log_two_tor(base_exp, pairing[4], key->b1, 1);
}

void decompress_two_tor(compressed_key *key, f2elm_t b_inv, f2elm_t result1, f2elm_t result2) {
    f2elm_t a, a24;
    f2elm_proj_t result1_proj, result2_proj;
    fp2copy751(key->a, a);
    compute_a24_mont(a, a24);

    full_point_proj_t Q1, Q2;
    get_generators_two_tor(a, a24, b_inv, Q1, Q2);

    comp_lin_comb(a, b_inv, a24, Q1, Q2, key->a1, key->a2, result1_proj);
    comp_lin_comb(a, b_inv, a24, Q1, Q2, key->b1, key->b2, result2_proj);

    get_affine(result1_proj, result1);
    get_affine(result2_proj, result2);
}

void decompress_three_tor(compressed_key *key, f2elm_t b_inv, f2elm_t result1, f2elm_t result2) {
    f2elm_t a, a24;
    f2elm_proj_t result1_proj, result2_proj;
    fp2copy751(key->a, a);
    compute_a24_mont(a, a24);

    full_point_proj_t Q1, Q2;
    get_generators_three_tor(a, a24, b_inv, Q1, Q2);

    comp_lin_comb(a, b_inv, a24, Q1, Q2, key->a1, key->a2, result1_proj);
    comp_lin_comb(a, b_inv, a24, Q1, Q2, key->b1, key->b2, result2_proj);

    get_affine(result1_proj, result1);
    get_affine(result2_proj, result2);
}

int is_whitespace(char c) {
    return (c == ' ') || (c == '\n') || (c == '\r');
}

void benchmark_two_tor() {
    FILE *f = fopen("testing_data/2_tor_tests", "r");
    int i, test_num;
    unsigned int c;
    unsigned long comp_time = 0;
    unsigned long decomp_time = 0;
    clock_t tmp1, tmp2;

    if (f == NULL) {
        fprintf(stderr, "ERROR: Failed to open testing_data/2_tor_tests for benchmarking\n");
        exit(1);
    }

    printf("Benchmarking (two torsion)");
    fflush(stdout);

    test_num = 0;
    do {
        while (((c = getc(f)) != EOF) && is_whitespace(c));
        if (c == EOF) break;
        ungetc(c, f);        

        ++test_num;
        printf(".");
        fflush(stdout);

        compressed_key key;
        f2elm_t a, x1_in, x2_in, x1_out, x2_out;

        for (i = 0; i < NWORDS_FIELD; i++) {
            fscanf(f, "%lx", &(a[0][i]));
        }

        for (i = 0; i < NWORDS_FIELD; i++) {
            fscanf(f, "%lx", &(a[1][i]));
        }

        for (i = 0; i < NWORDS_FIELD; i++) {
            fscanf(f, "%lx", &(x1_in[0][i]));
        }

        for (i = 0; i < NWORDS_FIELD; i++) {
            fscanf(f, "%lx", &(x1_in[1][i]));
        }

        for (i = 0; i < NWORDS_FIELD; i++) {
            fscanf(f, "%lx", &(x2_in[0][i]));
        }

        for (i = 0; i < NWORDS_FIELD; i++) {
            fscanf(f, "%lx", &(x2_in[1][i]));
        }


        tmp1 = clock();
        compress_two_tor(a, f2_one_mont, x1_in, x2_in, &key);
        tmp2 = clock();

        comp_time += (tmp2 - tmp1);

        tmp1 = clock();
        decompress_two_tor(&key, f2_one_mont, x1_out, x2_out);
        tmp2 = clock();

        decomp_time += (tmp2 - tmp1);

        if ((!is_f2elm_eq(x1_in, x1_out)) || (!is_f2elm_eq(x2_in, x2_out))) {
            fprintf(stderr, "ERROR: Failed test %i (two torsion)\n", test_num);
            exit(1);
        }
    } while (1);

    if (test_num == 0) {
        printf("ERROR: Could not find any two torsion tests\n");
        return;
    }

    printf("\n");
    printf("Benchmarking sucessful (two torsion)\n");
    printf("Average compression time: %li microseconds\n", comp_time/test_num);
    printf("Average decompression time: %li microseconds\n", decomp_time/test_num);
}


void benchmark_three_tor() {
    FILE *f = fopen("testing_data/3_tor_tests", "r");
    int i, test_num;
    unsigned int c;
    unsigned long comp_time = 0;
    unsigned long decomp_time = 0;
    clock_t tmp1, tmp2;

    if (f == NULL) {
        fprintf(stderr, "ERROR: Failed to open testing_data/3_tor_tests for benchmarking\n");
        exit(1);
    }

    printf("Benchmarking (three torsion)");
    fflush(stdout);

    test_num = 0;
    do {
        while (((c = getc(f)) != EOF) && is_whitespace(c));
        if (c == EOF) break;
        ungetc(c, f);        

        ++test_num;
        printf(".");
        fflush(stdout);

        compressed_key key;
        f2elm_t a, x1_in, x2_in, x1_out, x2_out;

        for (i = 0; i < NWORDS_FIELD; i++) {
            fscanf(f, "%lx", &(a[0][i]));
        }

        for (i = 0; i < NWORDS_FIELD; i++) {
            fscanf(f, "%lx", &(a[1][i]));
        }

        for (i = 0; i < NWORDS_FIELD; i++) {
            fscanf(f, "%lx", &(x1_in[0][i]));
        }

        for (i = 0; i < NWORDS_FIELD; i++) {
            fscanf(f, "%lx", &(x1_in[1][i]));
        }

        for (i = 0; i < NWORDS_FIELD; i++) {
            fscanf(f, "%lx", &(x2_in[0][i]));
        }

        for (i = 0; i < NWORDS_FIELD; i++) {
            fscanf(f, "%lx", &(x2_in[1][i]));
        }

        tmp1 = clock();
        compress_three_tor(a, f2_one_mont, x1_in, x2_in, &key);
        tmp2 = clock();

        comp_time += (tmp2 - tmp1);

        tmp1 = clock();
        decompress_three_tor(&key, f2_one_mont, x1_out, x2_out);
        tmp2 = clock();

        decomp_time += (tmp2 - tmp1);

        if ((!is_f2elm_eq(x1_in, x1_out)) || (!is_f2elm_eq(x2_in, x2_out))) {
            fprintf(stderr, "ERROR: Failed test %i (three torsion)\n", test_num);
            exit(1);
        }
    } while (1);

    if (test_num == 0) {
        printf("ERROR: Could not find any three torsion tests\n");
        return;
    }

    printf("\n");
    printf("Benchmarking sucessful (three torsion)\n");
    printf("Average compression time: %li microseconds\n", comp_time/test_num);
    printf("Average decompression time: %li microseconds\n", decomp_time/test_num);
}


int main(int argc, char *argv[]) {

    init_field_constants();

    benchmark_two_tor();
    printf("\n");
    benchmark_three_tor();

    return 0;
}
