/********************************************************************************************
* Author:   David Urbanik;  dburbani@uwaterloo.ca 
* Abstract: Finite field arithmetic for ARM64 using code modified from the original x86_64
*           and generic implementations by Microsoft.
*
*********************************************************************************************/

#include "../definitions.h"


// Global constants
extern const uint64_t p751[NWORDS_FIELD];
extern const uint64_t p751p1[NWORDS_FIELD]; 

#include <stdio.h>
#include <stdlib.h>


__inline void fpadd751(digit_t* a, digit_t* b, digit_t* c)
{ // Modular addition, c = a+b mod p751.
  // Inputs: a, b in [0, p751-1] 
  // Output: c in [0, p751-1]

    fpadd751_arm64_asm(a, b, c);
} 


__inline void fpsub751(digit_t* a, digit_t* b, digit_t* c)
{ // Modular subtraction, c = a-b mod p751.
  // Inputs: a, b in [0, p751-1] 
  // Output: c in [0, p751-1] 

    fpsub751_arm64_asm(a, b, c);
}


//  identical to x86_64 implementation
__inline void fpneg751(digit_t* a)
{ // Modular negation, a = -a mod p751.
  // Input/output: a in [0, p751-1] 
    unsigned int i, borrow = 0;

    for (i = 0; i < NWORDS_FIELD; i++) {
        SUBC(borrow, ((digit_t*)p751)[i], a[i], borrow, a[i]); 
    }
}


//  identical to x86_64 implementation
void fpdiv2_751(digit_t* a, digit_t* c)
{ // Modular division by two, c = a/2 mod p751.
  // Input : a in [0, p751-1] 
  // Output: c in [0, p751-1] 
    unsigned int i, carry = 0;
    digit_t mask;
        
    mask = 0 - (digit_t)(a[0] & 1);    // If a is odd compute a+p521
    for (i = 0; i < NWORDS_FIELD; i++) {
        ADDC(carry, a[i], ((digit_t*)p751)[i] & mask, carry, c[i]); 
    }

    mp_shiftr1(c, NWORDS_FIELD);
} 


void mp_mul_comba(digit_t* a, digit_t* b, digit_t* c, unsigned int nwords)
{ // Multiprecision comba multiply, c = a*b, where lng(a) = lng(b) = nwords.

//    UNREFERENCED_PARAMETER(nwords);

    fpmul751_arm64_asm(a, b, c);

//      mp_mul_karatsuba(a, b, c, nwords);
}



void rdc_mont(dfelm_t ma, felm_t mc)
{ // Optimized Montgomery reduction using comba and exploiting the special form of the prime p751.
  // mc = ma*mb*R^-1 mod p751, where ma,mb,mc in [0, p751-1] and R = 2^768.
  // ma and mb are assumed to be in Montgomery representation.
  
    rdc751_arm64_asm(ma, mc);
}
